@@ -1,16 +1,18 @@
 #!/usr/bin/env python
 # This file is part of Xpra.
 # Copyright (C) 2012 Serviware (Arthur Huillet, <ahuillet@serviware.com>)
-# Copyright (C) 2012-2018 Antoine Martin <antoine@devloop.org.uk>
+# Copyright (C) 2012-2018 Antoine Martin <antoine@xpra.org>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
 import sys
 import logging
-from xpra.util import envbool, csv
+
+from xpra.util import envbool, envint, csv
 from xpra.os_util import POSIX, OSX, WIN32, PYTHON3, bytestostr
 from xpra.log import Logger, CaptureHandler
-from xpra.client.gl.gl_drivers import WHITELIST, GREYLIST, VERSION_REQ, BLACKLIST
+from xpra.client.gl.gl_drivers import WHITELIST, GREYLIST, VERSION_REQ, BLACKLIST, OpenGLFatalError
+
 log = Logger("opengl")
 
 required_extensions = ["GL_ARB_texture_rectangle", "GL_ARB_vertex_program"]
@@ -33,12 +35,17 @@ CAN_DOUBLE_BUFFER = not PYTHON3
 DEFAULT_DOUBLE_BUFFERED = WIN32 or CAN_DOUBLE_BUFFER
 DOUBLE_BUFFERED = envbool("XPRA_OPENGL_DOUBLE_BUFFERED", DEFAULT_DOUBLE_BUFFERED)
 
+CRASH = envbool("XPRA_OPENGL_FORCE_CRASH", False)
+TIMEOUT = envint("XPRA_OPENGL_FORCE_TIMEOUT", 0)
+
 
 #by default, we raise an ImportError as soon as we find something missing:
 def raise_error(msg):
     raise ImportError(msg)
+def raise_fatal_error(msg):
+    raise OpenGLFatalError(msg)
 gl_check_error = raise_error
-
+gl_fatal_error = raise_fatal_error
 
 _version_warning_shown = False
 #support for memory views requires Python 2.7 and PyOpenGL 3.1
@@ -79,14 +86,37 @@ def check_functions(*functions):
         else:
             available.append(name)
     if len(missing)>0:
-        gl_check_error("some required OpenGL functions are not available: %s" % csv(missing))
+        raise_fatal_error("some required OpenGL functions are missing:\n%s" % csv(missing))
     else:
         log("All the required OpenGL functions are available: %s " % csv(available))
 
+def get_max_texture_size():
+    from OpenGL.GL import glGetInteger, GL_MAX_TEXTURE_SIZE
+    texture_size = glGetInteger(GL_MAX_TEXTURE_SIZE)
+    log("GL_MAX_TEXTURE_SIZE=%s", texture_size)
+    #this one may be missing?
+    rect_texture_size = texture_size
+    try:
+        from OpenGL.GL import GL_MAX_RECTANGLE_TEXTURE_SIZE
+        rect_texture_size = glGetInteger(GL_MAX_RECTANGLE_TEXTURE_SIZE)
+    except ImportError as e:
+        log("OpenGL: %s", e)
+        log("using GL_MAX_TEXTURE_SIZE=%s as default", texture_size)
+    else:
+        log("Texture size GL_MAX_RECTANGLE_TEXTURE_SIZE=%s", rect_texture_size)
+    return min(rect_texture_size, texture_size)
+
 
 def check_PyOpenGL_support(force_enable):
     props = {}
     try:
+        if CRASH:
+            import ctypes;
+            ctypes.string_at(0)
+            raise Exception("should have crashed!")
+        elif TIMEOUT>0:
+            import time
+            time.sleep(TIMEOUT)
         #log redirection:
         def redirect_log(logger_name):
             logger = logging.getLogger(logger_name)
@@ -108,14 +138,14 @@ def check_PyOpenGL_support(force_enable):
         from OpenGL.GL import glGetString, glGetInteger, glGetIntegerv
         gl_version_str = glGetString(GL_VERSION)
         if gl_version_str is None:
-            gl_check_error("OpenGL version is missing - cannot continue")
+            raise_fatal_error("OpenGL version is missing - cannot continue")
             return  {}
         gl_major = int(bytestostr(gl_version_str)[0])
         gl_minor = int(bytestostr(gl_version_str)[2])
         props["opengl"] = gl_major, gl_minor
         MIN_VERSION = (1,1)
         if (gl_major, gl_minor) < MIN_VERSION:
-            gl_check_error("OpenGL output requires version %s or greater, not %s.%s" %
+            raise_fatal_error("OpenGL output requires version %s or greater, not %s.%s" %
                               (".".join([str(x) for x in MIN_VERSION]), gl_major, gl_minor))
         else:
             log("found valid OpenGL version: %s.%s", gl_major, gl_minor)
@@ -136,12 +166,14 @@ def check_PyOpenGL_support(force_enable):
             global _version_warning_shown
             if not _version_warning_shown:
                 log.warn("Warning: version mismatch between PyOpenGL and PyOpenGL-accelerate")
+                log.warn(" %s vs %s", pyopengl_version, accel_version)
                 log.warn(" this may cause crashes")
                 _version_warning_shown = True
+                gl_check_error("PyOpenGL vs accelerate version mismatch: %s vs %s" % (pyopengl_version, accel_version))
         vsplit = pyopengl_version.split('.')
         #we now require PyOpenGL 3.1 or later
         if vsplit[:3]<['3','1'] and not force_enable:
-            gl_check_error("PyOpenGL version %s is too old and buggy" % pyopengl_version)
+            raise_fatal_error("PyOpenGL version %s is too old and buggy" % pyopengl_version)
             return {}
         props["zerocopy"] = bool(OpenGL_accelerate) and is_pyopengl_memoryview_safe(pyopengl_version, accel_version)
 
@@ -150,7 +182,7 @@ def check_PyOpenGL_support(force_enable):
         except:
             log("error querying extensions", exc_info=True)
             extensions = []
-            gl_check_error("OpenGL could not find the list of GL extensions - does the graphics driver support OpenGL?")
+            raise_fatal_error("OpenGL could not find the list of GL extensions - does the graphics driver support OpenGL?")
         log("OpenGL extensions found: %s", csv(extensions))
         props["extensions"] = extensions
 
@@ -217,7 +249,10 @@ def check_PyOpenGL_support(force_enable):
                 log.warn("Warning: %s '%s' is blacklisted!", *blacklisted)
                 log.warn(" force enabled by option")
             else:
-                gl_check_error("%s '%s' is blacklisted!" % (blacklisted))
+                if force_enable:
+                    log.warn("%s '%s' is blacklisted!" % (blacklisted))
+                else:
+                    raise_fatal_error("%s '%s' is blacklisted!" % (blacklisted))
         safe = bool(whitelisted) or not bool(blacklisted)
         if greylisted and not whitelisted:
             log.warn("Warning: %s '%s' is greylisted,", *greylisted)
@@ -241,11 +276,16 @@ def check_PyOpenGL_support(force_enable):
             glTexImage2D, \
             glMultiTexCoord2i, \
             glVertex2i, glEnd)
+        #check for framebuffer functions we need:
+        from OpenGL.GL.ARB.framebuffer_object import GL_FRAMEBUFFER, \
+            GL_COLOR_ATTACHMENT0, glGenFramebuffers, glBindFramebuffer, glFramebufferTexture2D
+        check_functions(GL_FRAMEBUFFER, \
+            GL_COLOR_ATTACHMENT0, glGenFramebuffers, glBindFramebuffer, glFramebufferTexture2D)
 
         glEnablei = None
         try:
             from OpenGL.GL import glEnablei
-        except:
+        except ImportError:
             pass
         if not bool(glEnablei):
             log.warn("OpenGL glEnablei is not available, disabling transparency")
@@ -253,15 +293,9 @@ def check_PyOpenGL_support(force_enable):
             GL_ALPHA_SUPPORTED = False
         props["transparency"] = GL_ALPHA_SUPPORTED
 
-        #check for framebuffer functions we need:
-        from OpenGL.GL.ARB.framebuffer_object import GL_FRAMEBUFFER, \
-            GL_COLOR_ATTACHMENT0, glGenFramebuffers, glBindFramebuffer, glFramebufferTexture2D
-        check_functions(GL_FRAMEBUFFER, \
-            GL_COLOR_ATTACHMENT0, glGenFramebuffers, glBindFramebuffer, glFramebufferTexture2D)
-
         for ext in required_extensions:
             if ext not in extensions:
-                gl_check_error("OpenGL driver lacks support for extension: %s" % ext)
+                raise_fatal_error("OpenGL driver lacks support for extension: %s" % ext)
             else:
                 log("Extension %s is present", ext)
 
@@ -269,13 +303,13 @@ def check_PyOpenGL_support(force_enable):
         #see http://www.opengl.org/registry/specs/ARB/fragment_program.txt
         from OpenGL.GL.ARB.fragment_program import glInitFragmentProgramARB
         if not glInitFragmentProgramARB():
-            gl_check_error("OpenGL output requires glInitFragmentProgramARB")
+            raise_fatal_error("OpenGL output requires glInitFragmentProgramARB")
         else:
             log("glInitFragmentProgramARB works")
 
         from OpenGL.GL.ARB.texture_rectangle import glInitTextureRectangleARB
         if not glInitTextureRectangleARB():
-            gl_check_error("OpenGL output requires glInitTextureRectangleARB")
+            raise_fatal_error("OpenGL output requires glInitTextureRectangleARB")
         else:
             log("glInitTextureRectangleARB works")
 
@@ -283,32 +317,13 @@ def check_PyOpenGL_support(force_enable):
             glBindProgramARB, glProgramStringARB
         check_functions(glGenProgramsARB, glDeleteProgramsARB, glBindProgramARB, glProgramStringARB)
 
-        try:
-            from OpenGL.GL import GL_MAX_TEXTURE_SIZE
-            texture_size = glGetInteger(GL_MAX_TEXTURE_SIZE)
-            #this one may be missing?
-            rect_texture_size = texture_size
-            try:
-                from OpenGL.GL import GL_MAX_RECTANGLE_TEXTURE_SIZE
-                rect_texture_size = glGetInteger(GL_MAX_RECTANGLE_TEXTURE_SIZE)
-            except ImportError as e:
-                log("OpenGL: %s", e)
-                log("using GL_MAX_TEXTURE_SIZE=%s as default", texture_size)
-        except Exception as e:
-            emsg = str(e)
-            if hasattr(e, "description"):
-                emsg = e.description
-            gl_check_error("unable to query max texture size: %s" % emsg)
-            return props
-
-        log("Texture size GL_MAX_RECTANGLE_TEXTURE_SIZE=%s, GL_MAX_TEXTURE_SIZE=%s", rect_texture_size, texture_size)
-        texture_size_limit = min(rect_texture_size, texture_size)
-        props["texture-size-limit"] = texture_size_limit
+        texture_size_limit = get_max_texture_size()
+        props["texture-size-limit"] = int(texture_size_limit)
 
         try:
             from OpenGL.GL import GL_MAX_VIEWPORT_DIMS
             v = glGetIntegerv(GL_MAX_VIEWPORT_DIMS)
-            max_viewport_dims = v[0], v[1]
+            max_viewport_dims = int(v[0]), int(v[1])
             assert max_viewport_dims[0]>=texture_size_limit and max_viewport_dims[1]>=texture_size_limit
             log("GL_MAX_VIEWPORT_DIMS=%s", max_viewport_dims)
         except ImportError as e:
@@ -342,7 +357,7 @@ def check_PyOpenGL_support(force_enable):
             if p>0:
                 format_handler = format_handler[:p]
                 missing_handlers.append(format_handler)
-        if len(missing_handlers)>0:
+        if missing_handlers:
             log.warn("PyOpenGL warning: missing array format handlers: %s", csv(missing_handlers))
 
         for x in elogger.handlers[0].records:
@@ -391,20 +406,21 @@ def main():
         if verbose:
             log.enable_debug()
         if POSIX and not OSX:
-            from xpra.x11.gtk_x11.gdk_display_source import init_display_source
-            init_display_source()
+            from xpra.x11.gtk_x11.gdk_display_source import init_gdk_display_source
+            init_gdk_display_source()
         force_enable = "-f" in sys.argv or "--force" in sys.argv
         from xpra.platform.gl_context import GLContext
         log("testing %s", GLContext)
         gl_context = GLContext()
         log("GLContext=%s", gl_context)
         #replace ImportError with a log message:
-        global gl_check_error
+        global gl_check_error, raise_fatal_error
         errors = []
         def log_error(msg):
             log.error("ERROR: %s", msg)
             errors.append(msg)
         gl_check_error = log_error
+        raise_fatal_error = log_error
         try:
             props = gl_context.check_support(force_enable)
         except Exception as e:
