@@ -1,6 +1,6 @@
 #!/usr/bin/env python
 # This file is part of Xpra.
-# Copyright (C) 2009-2018 Antoine Martin <antoine@devloop.org.uk>
+# Copyright (C) 2009-2019 Antoine Martin <antoine@xpra.org>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
@@ -12,43 +12,52 @@ This is a simple GUI for starting the xpra client.
 
 import os.path
 import sys
-import signal
 import traceback
 
-from xpra.gtk_common.gobject_compat import import_gtk, import_gdk, import_gobject, import_pango, import_glib
+from xpra.gtk_common.gobject_compat import (
+    import_gtk, import_gdk, import_gobject, import_pango, import_glib,
+    register_os_signals,
+    )
+from xpra.scripts.config import read_config, make_defaults_struct, validate_config, save_config
+from xpra.codecs.codec_constants import PREFERED_ENCODING_ORDER
+from xpra.gtk_common.quit import gtk_main_quit_really
+from xpra.gtk_common.gtk_util import (
+    gtk_main, add_close_accel, scaled_image, pixbuf_new_from_file, color_parse,
+    OptionMenu, choose_file, set_use_tray_workaround, window_defaults, imagebutton,
+    WIN_POS_CENTER, STATE_NORMAL,
+    DESTROY_WITH_PARENT, MESSAGE_INFO,  BUTTONS_CLOSE,
+    FILE_CHOOSER_ACTION_SAVE, FILE_CHOOSER_ACTION_OPEN,
+    )
+from xpra.util import DEFAULT_PORT, csv, repr_ellipsized
+from xpra.os_util import thread, WIN32, OSX, PYTHON3
+from xpra.client.gtk_base.gtk_tray_menu_base import (
+    make_min_auto_menu, make_encodingsmenu,
+    MIN_QUALITY_OPTIONS, QUALITY_OPTIONS, MIN_SPEED_OPTIONS, SPEED_OPTIONS,
+    )
+from xpra.gtk_common.about import about
+from xpra.scripts.main import (
+    connect_to, make_client, configure_network, is_local,
+    add_ssh_args, parse_ssh_string, add_ssh_proxy_args,
+    )
+from xpra.platform.paths import get_icon_dir
+from xpra.platform import get_username
+from xpra.log import Logger, enable_debug_for
+
+log = Logger("launcher")
+
 gobject = import_gobject()
-gobject.threads_init()
 glib = import_glib()
-glib.threads_init()
 gtk = import_gtk()
 gdk = import_gdk()
 pango = import_pango()
 
-from xpra.platform.gui import init as gui_init
-gui_init()
-
-from xpra.scripts.config import read_config, make_defaults_struct, validate_config, save_config
-from xpra.codecs.codec_constants import PREFERED_ENCODING_ORDER
-from xpra.gtk_common.quit import gtk_main_quit_really
-from xpra.gtk_common.gtk_util import gtk_main, add_close_accel, scaled_image, pixbuf_new_from_file, color_parse, \
-                                    OptionMenu, choose_file, set_use_tray_workaround, window_defaults, imagebutton, \
-                                    WIN_POS_CENTER, STATE_NORMAL, \
-                                    DESTROY_WITH_PARENT, MESSAGE_INFO,  BUTTONS_CLOSE, \
-                                    FILE_CHOOSER_ACTION_SAVE, FILE_CHOOSER_ACTION_OPEN
-from xpra.util import DEFAULT_PORT, csv, repr_ellipsized
-from xpra.os_util import thread, WIN32, OSX, PYTHON3
-from xpra.client.gtk_base.gtk_tray_menu_base import make_min_auto_menu, make_encodingsmenu, \
-                                    MIN_QUALITY_OPTIONS, QUALITY_OPTIONS, MIN_SPEED_OPTIONS, SPEED_OPTIONS
-from xpra.gtk_common.about import about
-from xpra.scripts.main import connect_to, make_client, configure_network, is_local, add_ssh_args, parse_ssh_string
-from xpra.platform.paths import get_icon_dir
-from xpra.platform import get_username
-from xpra.log import Logger, enable_debug_for
-log = Logger("launcher")
-
 #what we save in the config file:
-SAVED_FIELDS = ["username", "password", "host", "port", "mode", "ssh_port",
-                "encoding", "quality", "min-quality", "speed", "min-speed"]
+SAVED_FIELDS = [
+    "username", "password", "host", "port", "mode", "ssh_port",
+    "encoding", "quality", "min-quality", "speed", "min-speed",
+    "proxy_port", "proxy_username", "proxy_key", "proxy_password",
+    "proxy_host"
+    ]
 
 #options not normally found in xpra config file
 #but which can be present in a launcher config:
@@ -60,6 +69,11 @@ LAUNCHER_OPTION_TYPES ={
                         "mode"              : str,
                         "autoconnect"       : bool,
                         "ssh_port"          : int,
+                        "proxy_host"        : str,
+                        "proxy_port"        : int,
+                        "proxy_username"    : str,
+                        "proxy_password"    : str,
+                        "proxy_key"         : str,
                         }
 LAUNCHER_DEFAULTS = {
                         "host"              : "",
@@ -69,6 +83,11 @@ LAUNCHER_DEFAULTS = {
                         "mode"              : "tcp",    #tcp,ssh,..
                         "autoconnect"       : False,
                         "ssh_port"          : 22,
+                        "proxy_host"        : "",
+                        "proxy_port"        : 22,
+                        "proxy_username"    : get_username(),
+                        "proxy_password"    : "",
+                        "proxy_key"   : "",
                     }
 
 
@@ -105,7 +124,6 @@ def has_mdns():
 
 def noop(*args):
     log("noop%s", args)
-    pass
 
 
 class ApplicationWindow:
@@ -113,13 +131,14 @@ class ApplicationWindow:
     def __init__(self):
         # Default connection options
         self.config = make_defaults_struct(extras_defaults=LAUNCHER_DEFAULTS, extras_types=LAUNCHER_OPTION_TYPES, extras_validation=self.get_launcher_validation())
+        ssh_cmd = parse_ssh_string(self.config.ssh)[0].strip().lower()
+        self.is_putty = ssh_cmd.endswith("plink") or ssh_cmd.endswith("plink.exe")
+        self.is_paramiko = ssh_cmd=="paramiko"
         #TODO: the fixup does not belong here?
         from xpra.scripts.main import fixup_options
         fixup_options(self.config)
         #what we save by default:
         self.config_keys = set(SAVED_FIELDS)
-        def raise_exception(*args):
-            raise Exception(*args)
         self.client = None
         self.exit_launcher = False
         self.exit_code = None
@@ -127,6 +146,7 @@ class ApplicationWindow:
 
     def get_connection_modes(self):
         modes = ["ssh"]
+        modes.append("ssh -> ssh")
         try:
             import ssl
             assert ssl
@@ -190,7 +210,7 @@ class ApplicationWindow:
         self.bug_tool = None
         if icon_pixbuf:
             def bug(*_args):
-                if self.bug_tool==None:
+                if self.bug_tool is None:
                     from xpra.client.gtk_base.bug_report import BugReport
                     self.bug_tool = BugReport()
                     self.bug_tool.init(show_about=False)
@@ -202,7 +222,7 @@ class ApplicationWindow:
         self.mdns_gui = None
         if icon_pixbuf and has_mdns():
             def mdns(*_args):
-                if self.mdns_gui==None:
+                if self.mdns_gui is None:
                     from xpra.client.gtk_base.mdns_gui import mdns_sessions
                     self.mdns_gui = mdns_sessions(self.config)
                     def close_mdns():
@@ -221,26 +241,103 @@ class ApplicationWindow:
         vbox.pack_start(hbox)
 
         # Mode:
-        hbox = gtk.HBox(False, 20)
-        hbox.set_spacing(20)
-        hbox.pack_start(gtk.Label("Mode: "))
+        hbox = gtk.HBox(False, 5)
         self.mode_combo = gtk.combo_box_new_text()
         for x in self.get_connection_modes():
             self.mode_combo.append_text(x.upper())
         self.mode_combo.connect("changed", self.mode_changed)
-        hbox.pack_start(self.mode_combo)
-        vbox.pack_start(hbox)
+        hbox.pack_start(gtk.Label("Mode: "), False, False)
+        hbox.pack_start(self.mode_combo, False, False)
+        align_hbox = gtk.Alignment(xalign = .5)
+        align_hbox.add(hbox)
+        vbox.pack_start(align_hbox)
 
-        # Username@Host:Port
-        hbox = gtk.HBox(False, 0)
-        hbox.set_spacing(5)
+        # Username@Host:Port (ssh -> ssh, proxy)
+        vbox_proxy = gtk.VBox(False, 15)
+        hbox = gtk.HBox(False, 5)
+        self.proxy_vbox = vbox_proxy
+        self.proxy_username_entry = gtk.Entry()
+        self.proxy_username_entry.set_max_length(128)
+        self.proxy_username_entry.set_width_chars(16)
+        self.proxy_username_entry.connect("changed", self.validate)
+        self.proxy_username_entry.connect("activate", self.connect_clicked)
+        self.proxy_username_entry.set_tooltip_text("username")
+        self.proxy_host_entry = gtk.Entry()
+        self.proxy_host_entry.set_max_length(128)
+        self.proxy_host_entry.set_width_chars(24)
+        self.proxy_host_entry.connect("changed", self.validate)
+        self.proxy_host_entry.connect("activate", self.connect_clicked)
+        self.proxy_host_entry.set_tooltip_text("hostname")
+        self.proxy_port_entry = gtk.Entry()
+        self.proxy_port_entry.set_max_length(5)
+        self.proxy_port_entry.set_width_chars(5)
+        self.proxy_port_entry.connect("changed", self.validate)
+        self.proxy_port_entry.connect("activate", self.connect_clicked)
+        self.proxy_port_entry.set_tooltip_text("SSH port")
+        hbox.pack_start(gtk.Label("Proxy: "), False, False)
+        hbox.pack_start(self.proxy_username_entry, True, True)
+        hbox.pack_start(gtk.Label("@"), False, False)
+        hbox.pack_start(self.proxy_host_entry, True, True)
+        hbox.pack_start(self.proxy_port_entry, False, False)
+        vbox_proxy.pack_start(hbox)
+
+        # Password
+        hbox = gtk.HBox(False, 5)
+        self.proxy_password_hbox = hbox
+        self.proxy_password_entry = gtk.Entry()
+        self.proxy_password_entry.set_max_length(128)
+        self.proxy_password_entry.set_width_chars(30)
+        self.proxy_password_entry.set_text("")
+        self.proxy_password_entry.set_visibility(False)
+        self.proxy_password_entry.connect("changed", self.password_ok)
+        self.proxy_password_entry.connect("changed", self.validate)
+        self.proxy_password_entry.connect("activate", self.connect_clicked)
+        hbox.pack_start(gtk.Label("Proxy Password"), False, False)
+        hbox.pack_start(self.proxy_password_entry, True, True)
+        vbox_proxy.pack_start(hbox)
+
+        # Private key
+        hbox = gtk.HBox(False,  5)
+        self.pkey_hbox = hbox
+        self.proxy_key_label = gtk.Label("Proxy private key path (PPK):")
+        self.proxy_key_entry = gtk.Entry()
+        self.proxy_key_browse = gtk.Button("Browse")
+        self.proxy_key_browse.connect("clicked", self.proxy_key_browse_clicked)
+        hbox.pack_start(self.proxy_key_label, False, False)
+        hbox.pack_start(self.proxy_key_entry, True, True)
+        hbox.pack_start(self.proxy_key_browse, False, False)
+        vbox_proxy.pack_start(hbox)
+
+        # Check boxes
+        hbox = gtk.HBox(False, 5)
+        self.check_boxes_hbox = hbox
+        self.password_scb = gtk.CheckButton("Server password same as proxy")
+        self.password_scb.set_mode(True)
+        self.password_scb.set_active(True)
+        self.password_scb.connect("toggled", self.validate)
+        align_password_scb = gtk.Alignment(xalign = 1.0)
+        align_password_scb.add(self.password_scb)
+        self.username_scb = gtk.CheckButton("Server username same as proxy")
+        self.username_scb.set_mode(True)
+        self.username_scb.set_active(True)
+        self.username_scb.connect("toggled", self.validate)
+        align_username_scb = gtk.Alignment(xalign = 0.0)
+        align_username_scb.add(self.username_scb)
+        hbox.pack_start(align_username_scb, True, True)
+        hbox.pack_start(align_password_scb, True, True)
+        vbox_proxy.pack_start(hbox)
+
+        # coniditonal stuff that goes away for "normal" ssh
+        vbox.pack_start(vbox_proxy)
+
+        # Username@Host:Port (main)
+        hbox = gtk.HBox(False, 5)
         self.username_entry = gtk.Entry()
         self.username_entry.set_max_length(128)
         self.username_entry.set_width_chars(16)
         self.username_entry.connect("changed", self.validate)
         self.username_entry.connect("activate", self.connect_clicked)
         self.username_entry.set_tooltip_text("username")
-        self.username_label = gtk.Label("@")
         self.host_entry = gtk.Entry()
         self.host_entry.set_max_length(128)
         self.host_entry.set_width_chars(24)
@@ -259,18 +356,18 @@ class ApplicationWindow:
         self.port_entry.connect("changed", self.validate)
         self.port_entry.connect("activate", self.connect_clicked)
         self.port_entry.set_tooltip_text("port/display")
-
-        hbox.pack_start(self.username_entry)
-        hbox.pack_start(self.username_label)
-        hbox.pack_start(self.host_entry)
-        hbox.pack_start(self.ssh_port_entry)
-        hbox.pack_start(gtk.Label(":"))
-        hbox.pack_start(self.port_entry)
+        hbox.pack_start(gtk.Label("Server:"), False, False)
+        hbox.pack_start(self.username_entry, True, True)
+        hbox.pack_start(gtk.Label("@"), False, False)
+        hbox.pack_start(self.host_entry, True, True)
+        hbox.pack_start(self.ssh_port_entry, False, False)
+        hbox.pack_start(gtk.Label(":"), False, False)
+        hbox.pack_start(self.port_entry, False, False)
         vbox.pack_start(hbox)
 
         # Password
-        hbox = gtk.HBox(False, 0)
-        hbox.set_spacing(20)
+        hbox = gtk.HBox(False, 5)
+        self.password_hbox = hbox
         self.password_entry = gtk.Entry()
         self.password_entry.set_max_length(128)
         self.password_entry.set_width_chars(30)
@@ -278,14 +375,13 @@ class ApplicationWindow:
         self.password_entry.set_visibility(False)
         self.password_entry.connect("changed", self.password_ok)
         self.password_entry.connect("changed", self.validate)
-        self.password_label = gtk.Label("Password: ")
-        hbox.pack_start(self.password_label)
-        hbox.pack_start(self.password_entry)
+        self.password_entry.connect("activate", self.connect_clicked)
+        hbox.pack_start(gtk.Label("Server Password:"), False, False)
+        hbox.pack_start(self.password_entry, True, True)
         vbox.pack_start(hbox)
 
         #strict host key check for SSL and SSH
-        hbox = gtk.HBox(False, 0)
-        hbox.set_spacing(20)
+        hbox = gtk.HBox(False, 5)
         self.nostrict_host_check = gtk.CheckButton("Disable Strict Host Key Check")
         self.nostrict_host_check.set_active(False)
         al = gtk.Alignment(xalign=0.5, yalign=0.5, xscale=0.0, yscale=0)
@@ -411,20 +507,47 @@ class ApplicationWindow:
         gtk_main_quit_really()
 
     def validate(self, *args):
-        ssh = self.mode_combo.get_active_text()=="SSH"
+        mode = self.mode_combo.get_active_text().lower()
+        ssh = mode=="ssh"
+        sshtossh = mode=="ssh -> ssh"
         errs = []
         host = self.host_entry.get_text()
         errs.append((self.host_entry, not bool(host), "specify the host"))
-        if ssh:
+        if ssh or sshtossh:
             #validate ssh port:
             ssh_port = self.ssh_port_entry.get_text()
             try:
                 ssh_port = int(ssh_port)
-            except:
+            except ValueError:
                 ssh_port = -1
             errs.append((self.ssh_port_entry, ssh_port<0 or ssh_port>=2**16, "invalid SSH port number"))
+        if sshtossh:
+            #validate ssh port:
+            proxy_port = self.proxy_port_entry.get_text()
+            try:
+                proxy_port = int(proxy_port)
+            except ValueError:
+                proxy_port = -1
+            errs.append((self.proxy_port_entry, proxy_port<0 or proxy_port>=2**16, "invalid SSH port number"))
         port = self.port_entry.get_text()
-        if ssh and not port:
+        if sshtossh:
+            if self.password_scb.get_active():
+                self.password_entry.set_sensitive(False)
+                self.password_entry.set_text(self.proxy_password_entry.get_text())
+            else:
+                self.password_entry.set_sensitive(True)
+            if self.username_scb.get_active():
+                self.username_entry.set_sensitive(False)
+                self.username_entry.set_text(self.proxy_username_entry.get_text())
+            else:
+                self.username_entry.set_sensitive(True)
+            errs.append((self.proxy_host_entry, not bool(self.proxy_host_entry.get_text()), "specify the proxy host"))
+        # check username *after* the checkbox action
+        if ssh or sshtossh:
+            errs.append((self.username_entry, not bool(self.username_entry.get_text()), "specify username"))
+        if sshtossh:
+            errs.append((self.proxy_username_entry, not bool(self.proxy_username_entry.get_text()), "specify proxy username"))
+        if ssh or sshtossh and not port:
             port = 0        #port optional with ssh
         else:
             try:
@@ -460,14 +583,25 @@ class ApplicationWindow:
     def mode_changed(self, *_args):
         mode = self.mode_combo.get_active_text().lower()
         ssh = mode=="ssh"
-        if ssh:
+        sshtossh = mode=="ssh -> ssh"
+        if ssh or sshtossh:
             self.port_entry.set_tooltip_text("Display number (optional)")
             self.port_entry.set_text("")
-            self.ssh_port_entry.set_text("22")
             self.ssh_port_entry.show()
             self.password_entry.set_tooltip_text("SSH Password")
             self.username_entry.set_tooltip_text("SSH Username")
+            if ssh:
+                self.proxy_vbox.hide()
+                self.password_scb.hide()
+                self.password_entry.set_sensitive(True)
+                self.username_entry.set_sensitive(True)
+            if sshtossh:
+                self.proxy_vbox.show()
+                self.password_scb.show()
         else:
+            self.password_entry.set_sensitive(True)
+            self.username_entry.set_sensitive(True)
+            self.proxy_vbox.hide()
             self.ssh_port_entry.hide()
             self.ssh_port_entry.set_text("")
             port_str = self.port_entry.get_text()
@@ -479,23 +613,37 @@ class ApplicationWindow:
             if self.config.port>0:
                 self.port_entry.set_text("%s" % self.config.port)
         can_use_password = True
-        if ssh:
-            ssh_cmd = parse_ssh_string(self.config.ssh)[0].strip().lower()
-            is_putty = ssh_cmd.endswith("plink") or ssh_cmd.endswith("plink.exe")
-            is_paramiko = ssh_cmd=="paramiko"
-            if is_putty or is_paramiko:
+        sshpass = False
+        if ssh or sshtossh:
+            if not self.is_putty:
+                self.proxy_key_entry.set_text("OpenSSH/Paramiko use ~/.ssh")
+                self.proxy_key_entry.set_editable(False)
+                self.proxy_key_entry.set_sensitive(False)
+                self.proxy_key_browse.hide()
+            if self.is_paramiko or self.is_putty:
                 can_use_password = True
             else:
                 #we can also use password if sshpass is installed:
                 from xpra.platform.paths import get_sshpass_command
                 sshpass = get_sshpass_command()
                 can_use_password = bool(sshpass)
+                sshpass = bool(sshpass)
         if can_use_password:
-            self.password_label.show()
-            self.password_entry.show()
+            self.password_hbox.show()
+            if sshtossh:
+                self.proxy_password_hbox.show()
+                # sshpass cannot do different username/passwords for proxy and destination
+                if not sshpass:
+                    self.check_boxes_hbox.show()
+                else:
+                    self.check_boxes_hbox.hide()
+                    self.password_scb.set_active(True)
+                    self.username_scb.set_active(True)
         else:
-            self.password_label.hide()
-            self.password_entry.hide()
+            self.password_hbox.hide()
+            if sshtossh:
+                self.check_boxes_hbox.hide()
+                self.proxy_password_hbox.hide()
         self.validate()
         if mode=="ssl" or (mode=="ssh" and not WIN32):
             self.nostrict_host_check.show()
@@ -531,7 +679,11 @@ class ApplicationWindow:
     def reset_errors(self):
         self.set_sensitive(True)
         self.set_info_text("")
-        for widget in (self.info, self.password_entry, self.username_entry, self.host_entry, self.port_entry):
+        for widget in (
+            self.info, self.password_entry, self.username_entry, self.host_entry,
+            self.port_entry, self.proxy_password_entry, self.proxy_username_entry,
+            self.proxy_host_entry, self.proxy_port_entry, self.ssh_port_entry,
+            ):
             self.set_widget_fg_color(self.info, False)
             self.set_widget_bg_color(widget, False)
 
@@ -546,6 +698,21 @@ class ApplicationWindow:
     def set_sensitive(self, s):
         glib.idle_add(self.window.set_sensitive, s)
 
+    def choose_pkey_file(self, title, action, action_button, callback):
+        file_filter = gtk.FileFilter()
+        file_filter.set_name("All Files")
+        file_filter.add_pattern("*")
+        choose_file(self.window, title, action, action_button, callback, file_filter)
+
+    def proxy_key_browse_clicked(self, *args):
+        log("proxy_key_browse_clicked%s", args)
+        def do_choose(filename):
+            #make sure the file extension is .ppk
+            if os.path.splitext(filename)[-1]!=".ppk":
+                filename += ".ppk"
+            self.proxy_key_entry.set_text(filename)
+        self.choose_pkey_file("Choose SSH private key", FILE_CHOOSER_ACTION_OPEN, gtk.STOCK_OPEN, do_choose)
+
     def connect_clicked(self, *args):
         log("connect_clicked%s", args)
         self.update_options_from_gui()
@@ -591,7 +758,7 @@ class ApplicationWindow:
         #cooked vars used by connect_to
         params = {"type"    : self.config.mode}
         username = self.config.username
-        if self.config.mode=="ssh":
+        if self.config.mode=="ssh" or self.config.mode=="ssh -> ssh":
             if self.config.socket_dir:
                 params["socket_dir"] = self.config.socket_dir
             params["remote_xpra"] = self.config.remote_xpra
@@ -602,12 +769,9 @@ class ApplicationWindow:
             else:
                 params["display"] = "auto"
                 params["display_as_args"] = []
-            full_ssh = parse_ssh_string(self.config.ssh)
-            ssh_cmd = full_ssh[0].lower()
-            is_putty = ssh_cmd.endswith("plink") or ssh_cmd.endswith("plink.exe")
-            is_paramiko = ssh_cmd=="paramiko"
-            params["is_putty"] = is_putty
-            params["is_paramiko"] = is_paramiko
+            params["ssh"] = self.config.ssh
+            params["is_putty"] = self.is_putty
+            params["is_paramiko"] = self.is_paramiko
             password = self.config.password
             host = self.config.host
             upos = host.find("@")
@@ -622,11 +786,26 @@ class ApplicationWindow:
                     username = username[:ppos]
             if self.config.ssh_port and self.config.ssh_port!=22:
                 params["ssh-port"] = self.config.ssh_port
-            full_ssh += add_ssh_args(username, password, host, self.config.ssh_port, is_putty, is_paramiko)
+            ssh_cmd = parse_ssh_string(self.config.ssh)
+            ssh_cmd_0 = ssh_cmd[0].strip().lower()
+            self.is_putty = ssh_cmd_0.endswith("plink") or ssh_cmd_0.endswith("plink.exe")
+            self.is_paramiko = ssh_cmd_0 =="paramiko"
+            full_ssh = ssh_cmd[:]
+            full_ssh += add_ssh_args(username, password, host, self.config.ssh_port, self.is_putty, self.is_paramiko)
             if username:
                 params["username"] = username
             if self.nostrict_host_check.get_active():
                 full_ssh += ["-o", "StrictHostKeyChecking=no"]
+            if params["type"] == "ssh -> ssh":
+                params["type"] = "ssh"
+                params["proxy_host"] = self.config.proxy_host
+                params["proxy_port"] = self.config.proxy_port
+                params["proxy_username"] = self.config.proxy_username
+                params["proxy_password"] = self.config.proxy_password
+                full_ssh += add_ssh_proxy_args(self.config.proxy_username, self.config.proxy_password,
+                                               self.config.proxy_host, self.config.proxy_port,
+                                               self.config.proxy_key, ssh_cmd,
+                                               self.is_putty, self.is_paramiko)
             params["host"] = host
             params["local"] = is_local(self.config.host)
             params["full_ssh"] = full_ssh
@@ -689,7 +868,7 @@ class ApplicationWindow:
             log.error("failed to start client", exc_info=True)
             self.handle_exception(e)
 
-    def do_start_XpraClient(self, conn, display_desc={}):
+    def do_start_XpraClient(self, conn, display_desc):
         log("do_start_XpraClient(%s, %s) client=%s", conn, display_desc, self.client)
         self.client.encoding = self.config.encoding
         self.client.display_desc = display_desc
@@ -720,7 +899,7 @@ class ApplicationWindow:
 
         def warn_and_quit_override(exit_code, warning):
             log("warn_and_quit_override(%s, %s)", exit_code, warning)
-            if self.exit_code == None:
+            if self.exit_code is None:
                 self.exit_code = exit_code
             password_warning = warning.find("invalid password")>=0
             if password_warning:
@@ -734,7 +913,7 @@ class ApplicationWindow:
 
         def quit_override(exit_code):
             log("quit_override(%s)", exit_code)
-            if self.exit_code == None:
+            if self.exit_code is None:
                 self.exit_code = exit_code
             handle_client_quit(self.exit_code==0)
 
@@ -779,33 +958,45 @@ class ApplicationWindow:
 
 
     def update_options_from_gui(self):
-        def pint(v):
+        def pint(vstr):
             try:
-                return int(v)
+                return int(vstr)
             except ValueError:
                 return 0
         self.config.host = self.host_entry.get_text()
         self.config.ssh_port = pint(self.ssh_port_entry.get_text())
         self.config.port = pint(self.port_entry.get_text())
         self.config.username = self.username_entry.get_text()
+        self.config.password = self.password_entry.get_text()
+
+        self.config.proxy_host = self.proxy_host_entry.get_text()
+        self.config.proxy_port = pint(self.proxy_port_entry.get_text())
+        self.config.proxy_username = self.proxy_username_entry.get_text()
+        self.config.proxy_password = self.proxy_password_entry.get_text()
+        if self.is_putty:
+            self.config.proxy_key = self.proxy_key_entry.get_text()
+
         self.config.encoding = self.get_selected_encoding() or self.config.encoding
         mode_enc = self.mode_combo.get_active_text().lower()
         if mode_enc.startswith("tcp"):
             self.config.mode = "tcp"
             if mode_enc.find("aes")>0:
                 self.config.encryption = "AES"
-        elif mode_enc in ("ssl", "ssh", "ws", "wss"):
+        elif mode_enc in ("ssl", "ssh", "ws", "wss", "ssh -> ssh"):
             self.config.mode = mode_enc
             self.config.encryption = ""
-        self.config.password = self.password_entry.get_text()
-        log("update_options_from_gui() %s", (self.config.username, self.config.password, self.config.mode, self.config.encryption, self.config.host, self.config.port, self.config.ssh_port, self.config.encoding))
+        log("update_options_from_gui() %s",
+            (self.config.username, self.config.password, self.config.mode, self.config.encryption, self.config.host, self.config.port, self.config.ssh_port, self.config.encoding))
 
     def update_gui_from_config(self):
         #mode:
         mode = (self.config.mode or "").lower()
+        active = 0
         for i,e in enumerate(self.get_connection_modes()):
             if e.lower()==mode:
-                self.mode_combo.set_active(i)
+                active = i
+                break
+        self.mode_combo.set_active(active)
         if self.config.encoding and self.encoding_combo:
             index = self.encoding_combo.get_menu().encoding_to_index.get(self.config.encoding, -1)
             log("setting encoding combo to %s / %s", self.config.encoding, index)
@@ -818,20 +1009,27 @@ class ApplicationWindow:
         self.username_entry.set_text(self.config.username)
         self.password_entry.set_text(self.config.password)
         self.host_entry.set_text(self.config.host)
-        def get_port(v, default_port=""):
+        def get_port(vstr, default_port=""):
             try:
-                iport = int(v)
-                if iport>0 and iport<2**16:
+                iport = int(vstr)
+                if 0<iport<2**16:
                     return str(iport)
-            except:
+            except ValueError:
                 pass
             return str(default_port)
         dport = DEFAULT_PORT
-        if mode=="ssh":
+        if mode in ("ssh", "ssh -> ssh"):
             #not required, so don't specify one
             dport = ""
         self.port_entry.set_text(get_port(self.config.port, dport))
         self.ssh_port_entry.set_text(get_port(self.config.ssh_port))
+        #proxy bits:
+        self.proxy_host_entry.set_text(self.config.proxy_host)
+        self.proxy_port_entry.set_text(get_port(self.config.proxy_port))
+        self.proxy_username_entry.set_text(self.config.proxy_username)
+        self.proxy_password_entry.set_text(self.config.proxy_password)
+        if self.is_putty:
+            self.proxy_key_entry.set_text(self.config.proxy_key)
 
     def close_window(self, *_args):
         w = self.window
@@ -927,10 +1125,10 @@ def do_main(argv):
     from xpra.os_util import SIGNAMES
     from xpra.scripts.main import InitExit, InitInfo
     from xpra.gtk_common.quit import gtk_main_quit_on_fatal_exceptions_enable
-    gtk_main_quit_on_fatal_exceptions_enable()
+    from xpra.platform.gui import init as gui_init, ready as gui_ready
 
-    from xpra.platform.gui import ready as gui_ready
     gui_init()
+    gtk_main_quit_on_fatal_exceptions_enable()
     try:
         from xpra.scripts.parsing import parse_cmdline, fixup_debug_option
         options, args = parse_cmdline(argv)
@@ -953,20 +1151,12 @@ def do_main(argv):
 
     try:
         app = ApplicationWindow()
-        def app_signal(signum, _frame):
-            print("")
-            log("got signal %s" % SIGNAMES.get(signum, signum))
-            def show_signal():
-                app.show()
-                app.client.cleanup()
-                glib.timeout_add(1000, app.set_info_text, "got signal %s" % SIGNAMES.get(signum, signum))
-                glib.timeout_add(1000, app.set_info_color, True)
-            #call from UI thread:
-            glib.idle_add(show_signal)
-        if sys.version_info[0]<3:
-            #breaks GTK3..
-            signal.signal(signal.SIGINT, app_signal)
-        signal.signal(signal.SIGTERM, app_signal)
+        def handle_signal(signum):
+            app.show()
+            app.client.cleanup()
+            glib.timeout_add(1000, app.set_info_text, "got signal %s" % SIGNAMES.get(signum, signum))
+            glib.timeout_add(1000, app.set_info_color, True)
+        register_os_signals(handle_signal)
         has_file = len(args) == 1
         if has_file:
             app.update_options_from_file(args[0])
@@ -1023,5 +1213,4 @@ def do_main(argv):
 
 
 if __name__ == "__main__":
-    v = main(sys.argv)
-    sys.exit(v)
+    sys.exit(main(sys.argv))
