@@ -1,6 +1,6 @@
 # This file is part of Xpra.
 # Copyright (C) 2008 Nathaniel Smith <njs@pobox.com>
-# Copyright (C) 2012-2017 Antoine Martin <antoine@devloop.org.uk>
+# Copyright (C) 2012-2019 Antoine Martin <antoine@xpra.org>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
@@ -8,28 +8,35 @@ import os
 import struct
 import re
 
+from xpra.gtk_common.gobject_util import no_arg_signal, SIGNAL_RUN_LAST
+from xpra.gtk_common.gtk_util import (
+    GetClipboard, PROPERTY_CHANGE_MASK,
+    selection_owner_set, selection_add_target, selectiondata_get_selection,
+    selectiondata_get_target, selectiondata_get_data,
+    selectiondata_get_data_type, selectiondata_get_format,
+    selectiondata_set, clipboard_request_contents,
+    set_clipboard_data,
+    )
+from xpra.gtk_common.nested_main import NestedMainLoop
+from xpra.net.compression import Compressible
+from xpra.os_util import WIN32, POSIX, monotonic_time, strtobytes, bytestostr, hexstr, get_hex_uuid, is_X11, is_Wayland
+from xpra.util import csv, envint, envbool, repr_ellipsized, typedict, first_time
+from xpra.platform.features import CLIPBOARD_GREEDY, CLIPBOARDS as PLATFORM_CLIPBOARDS
 from xpra.gtk_common.gobject_compat import import_gobject, import_gtk, import_gdk, import_glib, is_gtk3
+from xpra.log import Logger
+
+log = Logger("clipboard")
+
 gobject = import_gobject()
 glib = import_glib()
 gtk = import_gtk()
 gdk = import_gdk()
 
-from xpra.log import Logger
-log = Logger("clipboard")
-
-from xpra.gtk_common.gobject_util import no_arg_signal, SIGNAL_RUN_LAST
-from xpra.gtk_common.gtk_util import GetClipboard, selection_owner_set, selection_add_target, selectiondata_get_selection, selectiondata_get_target, selectiondata_get_data, selectiondata_get_data_type, selectiondata_get_format, selectiondata_set, clipboard_request_contents, PROPERTY_CHANGE_MASK
-from xpra.gtk_common.nested_main import NestedMainLoop
-from xpra.net.compression import Compressible
-from xpra.os_util import WIN32, POSIX, monotonic_time, strtobytes, bytestostr, hexstr, get_hex_uuid
-from xpra.util import csv, envint, envbool, repr_ellipsized, typedict
-from xpra.platform.features import CLIPBOARD_GREEDY
-
-
-MIN_CLIPBOARD_COMPRESSION_SIZE = 512
+MIN_CLIPBOARD_COMPRESS_SIZE = envint("XPRA_MIN_CLIPBOARD_COMPRESS_SIZE", 512)
 MAX_CLIPBOARD_PACKET_SIZE = 4*1024*1024
+MAX_CLIPBOARD_RECEIVE_SIZE = envint("XPRA_MAX_CLIPBOARD_RECEIVE_SIZE", -1)
+MAX_CLIPBOARD_SEND_SIZE = envint("XPRA_MAX_CLIPBOARD_SEND_SIZE", -1)
 
-from xpra.platform.features import CLIPBOARDS as PLATFORM_CLIPBOARDS
 ALL_CLIPBOARDS = [strtobytes(x) for x in PLATFORM_CLIPBOARDS]
 CLIPBOARDS = PLATFORM_CLIPBOARDS
 CLIPBOARDS_ENV = os.environ.get("XPRA_CLIPBOARDS")
@@ -47,18 +54,19 @@ LOOP_PREFIX = os.environ.get("XPRA_CLIPBOARD_LOOP_PREFIX", "Xpra-Clipboard-Loop-
 
 _discard_target_strs_ = os.environ.get("XPRA_DISCARD_TARGETS")
 if _discard_target_strs_ is not None:
-    DISCARD_TARGETS = _discard_target_strs_.split(",")
+    _discard_targets = _discard_target_strs_.split(",")
 else:
     #default:
-    DISCARD_TARGETS = [
-        "^SAVE_TARGETS$",
-        "^COMPOUND_TEXT",
-        "^NeXT",
-        "^com\.apple\.",
-        "^CorePasteboardFlavorType",
-        "^dyn\."]
-log("DISCARD_TARGETS=%s", csv(DISCARD_TARGETS))
-DISCARD_TARGETS = [re.compile(x) for x in DISCARD_TARGETS]
+    _discard_targets = [
+        r"^SAVE_TARGETS$",
+        r"^COMPOUND_TEXT",
+        r"^NeXT",
+        r"^com\.apple\.",
+        r"^CorePasteboardFlavorType",
+        r"^dyn\.",
+        ]
+log("discard_targets=%s", csv(_discard_targets))
+DISCARD_TARGETS = tuple(re.compile(dt) for dt in _discard_targets)
 
 TEXT_TARGETS = ("UTF8_STRING", "TEXT", "STRING", "text/plain")
 
@@ -66,9 +74,9 @@ TRANSLATED_TARGETS = {
     "application/x-moz-nativehtml" : "UTF8_STRING"
     }
 
-sizeof_long = struct.calcsize('@L')
+sizeof_long = struct.calcsize(b'@L')
 assert sizeof_long in (4, 8), "struct.calcsize('@L')=%s" % sizeof_long
-sizeof_short = struct.calcsize('=H')
+sizeof_short = struct.calcsize(b'=H')
 assert sizeof_short == 2, "struct.calcsize('=H')=%s" % sizeof_short
 
 
@@ -98,24 +106,10 @@ def _filter_targets(targets):
     log("_filter_targets(%s)=%s", targets, f)
     return f
 
-
-def set_string(clipboard, thestring):
-    if is_gtk3():
-        #no other way?
-        clipboard.set_text(thestring, len(thestring))
-        return
-    value = [thestring]
-    def get_func(clipboard, selection, info, targets):
-        log("get_func%s value=%s", (clipboard, selection, info, targets), value[0])
-        s = value[0]
-        if s:
-            selection.set("STRING", 8, s)
-        else:
-            clipboard.clear()
-    def clear_func(*args):
-        log("clear_func%s value=%s", args, value[0])
-        value[0] = ""
-    clipboard.set_with_data([("STRING", 0, 0)], get_func, clear_func)
+#CARD32 can actually be 64-bits...
+CARD32_SIZE = sizeof_long*8
+def get_format_size(dformat):
+    return max(8, {32 : CARD32_SIZE}.get(dformat, dformat))
 
 
 class ClipboardProtocolHelperBase(object):
@@ -125,7 +119,10 @@ class ClipboardProtocolHelperBase(object):
         self.progress_cb = progress_cb
         self.can_send = d.boolget("can-send", True)
         self.can_receive = d.boolget("can-receive", True)
-        self.max_clipboard_packet_size = MAX_CLIPBOARD_PACKET_SIZE
+        self.max_clipboard_packet_size = d.intget("max-packet-size", MAX_CLIPBOARD_PACKET_SIZE)
+        self.max_clipboard_receive_size = d.intget("max-receive-size", MAX_CLIPBOARD_RECEIVE_SIZE)
+        self.max_clipboard_send_size = d.intget("max-send-size", MAX_CLIPBOARD_SEND_SIZE)
+        self.clipboard_contents_slice_fix = False
         self.disabled_by_loop = []
         self.filter_res = []
         filter_res = d.strlistget("filters")
@@ -150,8 +147,10 @@ class ClipboardProtocolHelperBase(object):
 
     def get_info(self):
         info = {
-                "type"      : str(self),
-                "max_size"  : self.max_clipboard_packet_size,
+                "type"      :       str(self).replace("ClipboardProtocolHelper", ""),
+                "max_size"  :       self.max_clipboard_packet_size,
+                "max_recv_size":    self.max_clipboard_receive_size,
+                "max_send_size":    self.max_clipboard_send_size,
                 "filters"   : [x.pattern for x in self.filter_res],
                 "requests"  : self._clipboard_request_counter,
                 "pending"   : tuple(self._clipboard_outstanding_requests.keys()),
@@ -165,7 +164,7 @@ class ClipboardProtocolHelperBase(object):
         return info
 
     def cleanup(self):
-        def nosend(*args):
+        def nosend(*_args):
             pass
         self.send = nosend
         for x in self._clipboard_proxies.values():
@@ -204,9 +203,9 @@ class ClipboardProtocolHelperBase(object):
             for selection, rvalue in uuids.items():
                 log("%s=%s", proxy._selection, value)
                 if rvalue==proxy._loop_uuid:
-                    set_string(clipboard, "")
+                    set_clipboard_data(clipboard, "")
                 if rvalue and value==rvalue:
-                    set_string(clipboard, "")
+                    set_clipboard_data(clipboard, "")
                     if selection==proxy._selection:
                         log.warn("Warning: loop detected for %s clipboard", selection)
                     else:
@@ -218,12 +217,22 @@ class ClipboardProtocolHelperBase(object):
                         if selection not in self.disabled_by_loop:
                             self.disabled_by_loop.append(selection)
 
-    def set_direction(self, can_send, can_receive):
+    def set_direction(self, can_send, can_receive, max_send_size=None, max_receive_size=None):
         self.can_send = can_send
         self.can_receive = can_receive
+        self.set_limits(max_send_size, max_receive_size)
         for proxy in self._clipboard_proxies.values():
             proxy.set_direction(can_send, can_receive)
 
+    def set_limits(self, max_send_size, max_receive_size):
+        if max_send_size is not None:
+            self.max_clipboard_send_size = max_send_size
+        if max_receive_size is not None:
+            self.max_clipboard_receive_size = max_receive_size
+
+    def set_clipboard_contents_slice_fix(self, v):
+        self.clipboard_contents_slice_fix = v
+
     def enable_selections(self, selections):
         #when clients first connect or later through the "clipboard-enable-selections" packet,
         #they can tell us which clipboard selections they want enabled
@@ -278,10 +287,16 @@ class ClipboardProtocolHelperBase(object):
         return  selection
 
     # Used by the client during startup:
+    def send_tokens(self, selections=()):
+        for selection in selections:
+            proxy = self._clipboard_proxies.get(selection)
+            if proxy:
+                proxy._have_token = False
+                self._send_clipboard_token_handler(proxy)
+
     def send_all_tokens(self):
-        for proxy in self._clipboard_proxies.values():
-            proxy._have_token = False
-            self._send_clipboard_token_handler(proxy)
+        self.send_tokens(CLIPBOARDS)
+
 
     def _process_clipboard_token(self, packet):
         selection = bytestostr(packet[1])
@@ -293,13 +308,13 @@ class ClipboardProtocolHelperBase(object):
             l = log
             if name in ALL_CLIPBOARDS:
                 l = log.warn
-            l("ignoring token for clipboard proxy name '%s' (no proxy)", name)
+            l("ignoring token for clipboard '%s' (no proxy)", name)
             return
         if not proxy.is_enabled():
             l = log
             if name not in self.disabled_by_loop:
                 l = log.warn
-            l("ignoring token for clipboard proxy name '%s' (disabled)", name)
+            l("ignoring token for disabled clipboard '%s'", name)
             return
         log("process clipboard token selection=%s, local clipboard name=%s, proxy=%s", selection, name, proxy)
         targets = None
@@ -376,7 +391,7 @@ class ClipboardProtocolHelperBase(object):
                 send_now = [x for x in targets if x in TEXT_TARGETS]
                 def send_targets_only():
                     send_token(rsel, targets)
-                if len(send_now)==0:
+                if not send_now:
                     send_targets_only()
                     return
                 target = send_now[0]
@@ -429,43 +444,48 @@ class ClipboardProtocolHelperBase(object):
                 return None, None
             #important note: on 64 bits, format=32 means 8 bytes, not 4
             #that's just the way it is...
-            binfmt = "@" + "L" * (len(data) // sizeof_long)
+            binfmt = b"@" + b"L" * (len(data) // sizeof_long)
             ints = struct.unpack(binfmt, data)
             return b"integers", ints
-        elif dformat == 16:
-            binfmt = "=" + "H" * (len(data) // sizeof_short)
+        if dformat == 16:
+            binfmt = b"=" + b"H" * (len(data) // sizeof_short)
             ints = struct.unpack(binfmt, data)
             return b"integers", ints
-        elif dformat == 8:
+        if dformat == 8:
             for x in self.filter_res:
                 if x.match(data):
                     log.warn("clipboard buffer contains blacklisted pattern '%s' and has been dropped!", x.pattern)
                     return None, None
             return b"bytes", data
-        else:
-            log.error("unhandled format %s for clipboard data type %s" % (dformat, dtype))
-            return None, None
+        log.error("unhandled format %s for clipboard data type %s" % (dformat, dtype))
+        return None, None
 
     def _munge_wire_selection_to_raw(self, encoding, dtype, dformat, data):
-        log("wire selection to raw, encoding=%s, type=%s, format=%s, len(data)=%s", encoding, dtype, dformat, len(data or ""))
+        log("wire selection to raw, encoding=%s, type=%s, format=%s, len(data)=%s",
+            encoding, dtype, dformat, len(data or ""))
+        if self.max_clipboard_receive_size > 0:
+            max_recv_datalen = self.max_clipboard_receive_size * 8 // get_format_size(dformat)
+            if len(data) > max_recv_datalen:
+                olen = len(data)
+                data = data[:max_recv_datalen]
+                log.info("Data copied out truncated because of clipboard policy %d to %d", olen, max_recv_datalen)
         if encoding == b"bytes":
             return data
-        elif encoding == b"integers":
-            if len(data or "")==0:
+        if encoding == b"integers":
+            if not data:
                 return ""
             if dformat == 32:
-                format_char = "L"
+                format_char = b"L"
             elif dformat == 16:
-                format_char = "H"
+                format_char = b"H"
             elif dformat == 8:
-                format_char = "B"
+                format_char = b"B"
             else:
                 raise Exception("unknown encoding format: %s" % dformat)
-            fstr = "@" + format_char * len(data)
+            fstr = b"@" + format_char * len(data)
             log("struct.pack(%s, %s)", fstr, data)
             return struct.pack(fstr, *data)
-        else:
-            raise Exception("unhanled encoding: %s" % encoding)
+        raise Exception("unhanled encoding: %s" % encoding)
 
     def _process_clipboard_request(self, packet):
         request_id, selection, target = packet[1:4]
@@ -476,7 +496,8 @@ class ClipboardProtocolHelperBase(object):
             no_contents()
             return
         name = self.remote_to_local(bytestostr(selection))
-        log("process clipboard request, request_id=%s, selection=%s, local name=%s, target=%s", request_id, selection, name, target)
+        log("process clipboard request, request_id=%s, selection=%s, local name=%s, target=%s",
+            request_id, selection, name, target)
         proxy = self._clipboard_proxies.get(name)
         if proxy is None:
             #err, we were asked about a clipboard we don't handle..
@@ -503,6 +524,13 @@ class ClipboardProtocolHelperBase(object):
             if dtype is None or data is None or (dformat==0 and data==b""):
                 no_contents()
                 return
+            log("perform clipboard limit checking - datasize - %d, %d", len(data), self.max_clipboard_send_size)
+            truncated = 0
+            if self.max_clipboard_send_size > 0:
+                max_send_datalen = self.max_clipboard_send_size * 8 // get_format_size(dformat)
+                if len(data) > max_send_datalen:
+                    truncated = len(data) - max_send_datalen
+                    data = data[:max_send_datalen]
             munged = self._munge_raw_selection_to_wire(target, dtype, dformat, data)
             wire_encoding, wire_data = munged
             log("clipboard raw -> wire: %r -> %r", (dtype, dformat, data), munged)
@@ -511,8 +539,12 @@ class ClipboardProtocolHelperBase(object):
                 return
             wire_data = self._may_compress(dtype, dformat, wire_data)
             if wire_data is not None:
-                self.send("clipboard-contents", request_id, selection,
-                       dtype, dformat, wire_encoding, wire_data)
+                packet = ["clipboard-contents", request_id, selection,
+                        dtype, dformat, wire_encoding, wire_data]
+                if self.clipboard_contents_slice_fix:
+                    #sending the extra argument requires the fix
+                    packet.append(truncated)
+                self.send(*packet)
         proxy.get_contents(target, got_contents)
 
     def _may_compress(self, dtype, dformat, wire_data):
@@ -520,12 +552,12 @@ class ClipboardProtocolHelperBase(object):
             log.warn("Warning: clipboard contents are too big and have not been sent")
             log.warn(" %s compressed bytes dropped (maximum is %s)", len(wire_data), self.max_clipboard_packet_size)
             return  None
-        if type(wire_data)==str and len(wire_data)>=MIN_CLIPBOARD_COMPRESSION_SIZE:
+        if isinstance(wire_data, (str, bytes)) and len(wire_data)>=MIN_CLIPBOARD_COMPRESS_SIZE:
             return Compressible("clipboard: %s / %s" % (dtype, dformat), wire_data)
         return wire_data
 
     def _process_clipboard_contents(self, packet):
-        request_id, selection, dtype, dformat, wire_encoding, wire_data = packet[1:8]
+        request_id, selection, dtype, dformat, wire_encoding, wire_data = packet[1:7]
         log("process clipboard contents, selection=%s, type=%s, format=%s", selection, dtype, dformat)
         raw_data = self._munge_wire_selection_to_raw(wire_encoding, dtype, dformat, wire_data)
         log("clipboard wire -> raw: %r -> %r", (dtype, dformat, wire_encoding, wire_data), raw_data)
@@ -555,7 +587,7 @@ class ClipboardProtocolHelperBase(object):
         handler = self._packet_handlers.get(packet_type)
         log("process clipboard handler(%s)=%s", packet_type, handler)
         if handler:
-            self._packet_handlers[packet_type](packet)
+            handler(packet)
         else:
             log.warn("Warning: no clipboard packet handler for '%s'", packet_type)
 
@@ -608,11 +640,14 @@ class ClipboardProxy(gtk.Invisible):
         self._request_contents_events = 0
         self._last_targets = ()
 
-        try:
-            from xpra.x11.gtk_x11.prop import prop_get
-            self.prop_get = prop_get
-        except ImportError:
-            self.prop_get = None
+        if is_X11():
+            try:
+                from xpra.x11.gtk_x11.prop import prop_get
+                self.prop_get = prop_get
+            except ImportError as e:
+                log.warn("Warning: limited support for clipboard properties")
+                log.warn(" %s", e)
+                self.prop_get = None
 
         self._loop_uuid = ""
         self._clipboard.connect("owner-change", self.do_owner_changed)
@@ -620,7 +655,7 @@ class ClipboardProxy(gtk.Invisible):
     def init_uuid(self):
         self._loop_uuid = LOOP_PREFIX+get_hex_uuid()
         log("init_uuid() %s uuid=%s", self._selection, self._loop_uuid)
-        set_string(self._clipboard, self._loop_uuid)
+        set_clipboard_data(self._clipboard, "")
 
     def set_direction(self, can_send, can_receive):
         self._can_send = can_send
@@ -670,7 +705,10 @@ class ClipboardProxy(gtk.Invisible):
     def do_owner_changed(self, *_args):
         #an application on our side owns the clipboard selection
         #(they are ready to provide something via the clipboard)
-        log("clipboard: %s owner_changed, enabled=%s, can-send=%s, can-receive=%s, have_token=%s, greedy_client=%s, block_owner_change=%s", bytestostr(self._selection), self._enabled, self._can_send, self._can_receive, self._have_token, self._greedy_client, self._block_owner_change)
+        log("clipboard: %s owner_changed, enabled=%s, "+
+            "can-send=%s, can-receive=%s, have_token=%s, greedy_client=%s, block_owner_change=%s",
+            bytestostr(self._selection), self._enabled, self._can_send, self._can_receive,
+            self._have_token, self._greedy_client, self._block_owner_change)
         if not self._enabled or self._block_owner_change:
             return
         if self._have_token or (self._greedy_client and self._can_send):
@@ -746,7 +784,7 @@ class ClipboardProxy(gtk.Invisible):
         # assert that we can handle the requested sort of target. Fortunately,
         # Xpra never needs to request the clipboard when it owns it, so that's
         # okay.
-        assert str(event.selection) == self._selection, "selection does not match: expected %s but got %s" % (event.selection, self._selection)
+        assert str(event.selection) == self._selection, "expected %s but got %s" % (event.selection, self._selection)
         target = str(event.target)
         if target == "TIMESTAMP":
             pass
@@ -762,8 +800,9 @@ class ClipboardProxy(gtk.Invisible):
                 for t in targets:
                     selection_add_target(self, self._selection, t, 0)
         else:
-            log("target for %s: %r", self._selection, target)
-            selection_add_target(self, self._selection, target, 0)
+            if not must_discard(target):
+                log("target for %s: %r", self._selection, target)
+                selection_add_target(self, self._selection, target, 0)
         log("do_selection_request_event(%s) target=%s, selection=%s", event, target, self._selection)
         gtk.Invisible.do_selection_request_event(self, event)
 
@@ -778,11 +817,23 @@ class ClipboardProxy(gtk.Invisible):
         if not self._enabled or not self._can_receive:
             nodata()
             return
+        if not self._have_token:
+            try:
+                return gtk.Invisible.do_selection_get(self, selection_data, info, time)
+            except Exception as e:
+                log("gtk.Invisible.do_selection_get", exc_info=True)
+                if first_time("selection-%s-not-implemented" % self._selection):
+                    log.warn("Warning: limited clipboard support for %s", self._selection)
+                    if is_Wayland():
+                        log.warn(" looks like a Wayland implementation limitation")
+                    log.warn(" %s", e)
+                nodata()
+                return
         selection = selectiondata_get_selection(selection_data)
         target = selectiondata_get_target(selection_data)
         log("do_selection_get(%s, %s, %s) selection=%s", selection_data, info, time, selection)
         self._selection_get_events += 1
-        assert str(selection) == self._selection, "selection does not match: expected %s but got %s" % (selection, self._selection)
+        assert str(selection) == self._selection, "expected %s but got %s" % (selection, self._selection)
         self._request_contents_events += 1
         result = self.emit("get-clipboard-from-remote", self._selection, target)
         if result is None or result["type"] is None:
@@ -807,12 +858,13 @@ class ClipboardProxy(gtk.Invisible):
 
     def do_selection_clear_event(self, event):
         # Someone else on our side has the selection
-        log("do_selection_clear_event(%s) have_token=%s, block_owner_change=%s selection=%s", event, self._have_token, self._block_owner_change, self._selection)
+        log("do_selection_clear_event(%s) have_token=%s, block_owner_change=%s selection=%s",
+            event, self._have_token, self._block_owner_change, self._selection)
         self._selection_clear_events += 1
         if self._enabled and not self._block_owner_change:
             #if greedy_client is set, do_owner_changed will fire the token
             #so don't bother sending it now (same if we don't have it)
-            send = ((self._greedy_client and not self._block_owner_change) or self._have_token)
+            send = self._have_token or (self._greedy_client and not self._block_owner_change)
             self._have_token = False
 
             # Emit a signal -> send a note to the other side saying "hey its
@@ -828,7 +880,8 @@ class ClipboardProxy(gtk.Invisible):
         if not self._enabled:
             return
         self._got_token_events += 1
-        log("got token, selection=%s, targets=%s, target data=%s, claim=%s, can-receive=%s", self._selection, targets, target_data, claim, self._can_receive)
+        log("got token, selection=%s, targets=%s, target data=%s, claim=%s, can-receive=%s",
+            self._selection, targets, target_data, claim, self._can_receive)
         if self._greedy_client or CLIPBOARD_GREEDY:
             self._block_owner_change = True
             #re-enable the flag via idle_add so events like do_owner_changed
@@ -844,7 +897,7 @@ class ClipboardProxy(gtk.Invisible):
                     if text_target in target_data:
                         text_data = target_data.get(text_target)
                         log("clipboard %s set to '%s'", self._selection, repr_ellipsized(text_data))
-                        set_string(self._clipboard, text_data)
+                        set_clipboard_data(self._clipboard, text_data, text_target)
         if not claim:
             log("token packet without claim, not setting the token flag")
             #the other end is just telling us to send the token again next time something changes,
@@ -877,14 +930,15 @@ class ClipboardProxy(gtk.Invisible):
     # This function is called by the xpra core when the peer has requested the
     # contents of this clipboard:
     def get_contents(self, target, cb):
-        log("get_contents(%s,%s) selection=%s, enabled=%s, can-send=%s", target, cb, self._selection, self._enabled, self._can_send)
+        log("get_contents(%s, %s) selection=%s, enabled=%s, can-send=%s",
+            target, cb, self._selection, self._enabled, self._can_send)
         if not self._enabled or not self._can_send:
             cb(None, None, None)
             return
         self._get_contents_events += 1
         if self._have_token:
-            log.warn("Our peer requested the contents of the clipboard, but "
-                     + "*I* thought *they* had it... weird.")
+            log.warn("Warning: our peer requested the contents of the clipboard,")
+            log.warn(" but *I* thought *they* had it... weird.")
             cb(None, None, None)
             return
         if target=="TARGETS":
