@@ -1,5 +1,5 @@
 # This file is part of Xpra.
-# Copyright (C) 2011-2014 Antoine Martin <antoine@devloop.org.uk>
+# Copyright (C) 2011-2019 Antoine Martin <antoine@xpra.org>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
@@ -16,10 +16,13 @@ FLAGS_YAML      = 0x4
 ZLIB_FLAG       = 0x0       #assume zlib if no other compression flag is set
 LZ4_FLAG        = 0x10
 LZO_FLAG        = 0x20
-FLAGS_NOHEADER  = 0x40
+FLAGS_NOHEADER  = 0x10000   #never encoded, so we can use a value bigger than a byte
 
 
 _header_unpack_struct = struct.Struct(b'!cBBBL')
+HEADER_SIZE = _header_unpack_struct.size
+assert HEADER_SIZE==8
+
 def unpack_header(buf):
     return _header_unpack_struct.unpack_from(buf)
 
