@@ -1,23 +1,26 @@
 # -*- coding: utf-8 -*-
 # This file is part of Xpra.
-# Copyright (C) 2010-2018 Antoine Martin <antoine@devloop.org.uk>
+# Copyright (C) 2010-2018 Antoine Martin <antoine@xpra.org>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
+#pylint: disable-msg=E1101
 
 import os.path
 
-from xpra.log import Logger
-log = Logger("command")
-
 from xpra.util import parse_scaling_value, from0to100
 from xpra.server.control_command import ArgsControlCommand, ControlError
 from xpra.os_util import load_binary_file
 from xpra.util import csv
 from xpra.scripts.config import parse_bool, FALSE_OPTIONS, TRUE_OPTIONS
 from xpra.server.mixins.stub_server_mixin import StubServerMixin
+from xpra.log import Logger
 
-TOGGLE_FEATURES = ("bell", "randr", "cursors", "notifications", "dbus-proxy", "clipboard",
-                   "start-new-commands", "client-shutdown", "webcam", )
+log = Logger("command")
+
+TOGGLE_FEATURES = (
+    "bell", "randr", "cursors", "notifications", "dbus-proxy", "clipboard",
+    "start-new-commands", "client-shutdown", "webcam",
+    )
 
 
 """
@@ -33,10 +36,9 @@ class ServerBaseControlCommands(StubServerMixin):
         def parse_boolean_value(v):
             if str(v).lower() in TRUE_OPTIONS:
                 return True
-            elif str(v).lower() in FALSE_OPTIONS:
+            if str(v).lower() in FALSE_OPTIONS:
                 return False
-            else:
-                raise ControlError("a boolean is required, not %s" % v)
+            raise ControlError("a boolean is required, not %s" % v)
         def parse_4intlist(v):
             if not v:
                 return []
@@ -75,6 +77,7 @@ class ServerBaseControlCommands(StubServerMixin):
             ArgsControlCommand("compression",           "sets the packet compressor",       min_args=1, max_args=1),
             ArgsControlCommand("encoder",               "sets the packet encoder",          min_args=1, max_args=1),
             ArgsControlCommand("clipboard-direction",   "restrict clipboard transfers",     min_args=1, max_args=1),
+            ArgsControlCommand("clipboard-limits",      "restrict clipboard transfers size", min_args=2, max_args=2, validation=[int, int]),
             ArgsControlCommand("set-lock",              "modify the lock attribute",        min_args=1, max_args=1),
             ArgsControlCommand("set-sharing",           "modify the sharing attribute",     min_args=1, max_args=1),
             ArgsControlCommand("set-ui-driver",         "set the client connection driving the session", min_args=1, max_args=1),
@@ -106,7 +109,7 @@ class ServerBaseControlCommands(StubServerMixin):
         #encoding bits:
         for name in ("quality", "min-quality", "speed", "min-speed"):
             fn = getattr(self, "control_command_%s" % name.replace("-", "_"))
-            self.control_commands[name] = ArgsControlCommand(name, "set encoding %s (from 0 to 100)" % name, run=fn, min_args=1, max_args=1, validation=[from0to100])
+            self.control_commands[name] = ArgsControlCommand(name, "set encoding %s (from 0 to 100)" % name, run=fn, min_args=1, validation=[from0to100])
 
 
     #########################################
@@ -447,12 +450,21 @@ class ServerBaseControlCommands(StubServerMixin):
         self.clipboard_direction = direction
         can_send = direction in ("to-server", "both")
         can_receive = direction in ("to-client", "both")
-        self._clipboard_helper.set_direction(can_send, can_receive)
+        ch.set_direction(can_send, can_receive)
         msg = "clipboard direction set to '%s'" % direction
         log(msg)
         self.setting_changed("clipboard_direction", direction)
         return msg
 
+    def control_command_clipboard_limits(self, max_send, max_recv, *_args):
+        ch = self._clipboard_helper
+        assert self.clipboard and ch
+        ch.set_limits(max_send, max_recv)
+        msg = "clipboard send limit set to %d, recv limit set to %d (single copy/paste)" % (max_send, max_recv)
+        log(msg)
+        self.setting_changed("clipboard-limits", {'send': max_send, 'recv': max_recv})
+        return msg
+
     def _control_video_subregions_from_wid(self, wid):
         if wid not in self._id_to_window:
             raise ControlError("invalid window %i" % wid)
