@@ -1,35 +1,28 @@
 # -*- coding: utf-8 -*-
 # This file is part of Xpra.
 # Copyright (C) 2011 Serviware (Arthur Huillet, <ahuillet@serviware.com>)
-# Copyright (C) 2010-2018 Antoine Martin <antoine@devloop.org.uk>
+# Copyright (C) 2010-2018 Antoine Martin <antoine@xpra.org>
 # Copyright (C) 2008 Nathaniel Smith <njs@pobox.com>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
+from threading import Thread
 from time import sleep
 
-from xpra.log import Logger
-log = Logger("server")
-netlog = Logger("network")
-httplog = Logger("http")
-timeoutlog = Logger("timeout")
-screenlog = Logger("screen")
-
 from xpra.server.server_core import ServerCore, get_thread_info
 from xpra.server.mixins.server_base_controlcommands import ServerBaseControlCommands
-
-from xpra.os_util import thread, monotonic_time, bytestostr, WIN32, PYTHON3
-from xpra.util import typedict, flatten_dict, updict, merge_dicts, envbool, \
-    SERVER_EXIT, SERVER_ERROR, SERVER_SHUTDOWN, DETACH_REQUEST, NEW_CLIENT, DONE, SESSION_BUSY
+from xpra.os_util import monotonic_time, bytestostr, strtobytes, WIN32, PYTHON3
+from xpra.util import (
+    typedict, flatten_dict, updict, merge_dicts, envbool, envint,
+    SERVER_EXIT, SERVER_ERROR, SERVER_SHUTDOWN, DETACH_REQUEST,
+    NEW_CLIENT, DONE, SESSION_BUSY,
+    )
 from xpra.net.bytestreams import set_socket_timeout
+from xpra.server import server_features
 from xpra.server import EXITING_CODE
-
-
-CLIENT_CAN_SHUTDOWN = envbool("XPRA_CLIENT_CAN_SHUTDOWN", True)
-
+from xpra.log import Logger
 
 SERVER_BASES = [ServerCore, ServerBaseControlCommands]
-from xpra.server import server_features
 if server_features.notifications:
     from xpra.server.mixins.notification_forwarder import NotificationForwarder
     SERVER_BASES.append(NotificationForwarder)
@@ -74,8 +67,18 @@ if server_features.windows:
     SERVER_BASES.append(WindowServer)
 SERVER_BASES = tuple(SERVER_BASES)
 ServerBaseClass = type('ServerBaseClass', SERVER_BASES, {})
+
+log = Logger("server")
+netlog = Logger("network")
+httplog = Logger("http")
+timeoutlog = Logger("timeout")
+screenlog = Logger("screen")
+
 log("ServerBaseClass%s", SERVER_BASES)
 
+CLIENT_CAN_SHUTDOWN = envbool("XPRA_CLIENT_CAN_SHUTDOWN", True)
+INIT_THREAD_TIMEOUT = envint("XPRA_INIT_THREAD_TIMEOUT", 10)
+
 
 """
 This is the base class for seamless and desktop servers. (not proxy servers)
@@ -99,10 +102,10 @@ class ServerBase(ServerBaseClass):
         self.ui_driver = None
         self.sharing = None
         self.lock = None
+        self.init_thread = None
 
         self.idle_timeout = 0
         #duplicated from Server Source...
-        self.mem_bytes = 0
         self.client_shutdown = CLIENT_CAN_SHUTDOWN
         self.mp3_stream_check_timer = None
 
@@ -146,16 +149,27 @@ class ServerBase(ServerBaseClass):
         log("starting component init")
         for c in SERVER_BASES:
             c.setup(self)
-        thread.start_new_thread(self.threaded_init, ())
+        self.init_thread = Thread(target=self.threaded_init)
+        self.init_thread.start()
 
     def threaded_init(self):
         log("threaded_init() start")
-        sleep(0.1)
+        from xpra.platform import threaded_server_init
+        threaded_server_init()
         for c in SERVER_BASES:
             if c!=ServerCore:
                 c.threaded_setup(self)
         log("threaded_init() end")
 
+    def wait_for_threaded_init(self):
+        assert self.init_thread
+        log("wait_for_threaded_init() %s.is_alive()=%s", self.init_thread, self.init_thread.is_alive())
+        if self.init_thread.is_alive():
+            log.info("waiting for video encoders initialization")
+            self.init_thread.join(INIT_THREAD_TIMEOUT)
+            if self.init_thread.is_alive():
+                log.warn("Warning: initialization thread is still active")
+
 
     def server_is_ready(self):
         ServerCore.server_is_ready(self)
@@ -316,13 +330,14 @@ class ServerBase(ServerBaseClass):
 
         def drop_client(reason="unknown", *args):
             self.disconnect_client(proto, reason, *args)
-        from xpra.server.source.client_connection import ClientConnection
-        ss = ClientConnection(proto, drop_client,
-                          self.session_name, self,
-                          self.idle_add, self.timeout_add, self.source_remove,
-                          self.setting_changed,
-                          self._socket_dir, self.unix_socket_paths, not is_request, self.bandwidth_limit, self.bandwidth_detection,
-                          )
+        cc_class = self.get_client_connection_class()
+        ss = cc_class(proto, drop_client,
+                      self.session_name, self,
+                      self.idle_add, self.timeout_add, self.source_remove,
+                      self.setting_changed,
+                      self._socket_dir, self.unix_socket_paths, not is_request,
+                      self.bandwidth_limit, self.bandwidth_detection,
+                      )
         log("process_hello clientconnection=%s", ss)
         try:
             ss.parse_hello(c)
@@ -334,42 +349,52 @@ class ServerBase(ServerBaseClass):
         self._server_sources[proto] = ss
         #process ui half in ui thread:
         send_ui = ui_client and not is_request
-        self.idle_add(self.parse_hello_ui, ss, c, auth_caps, send_ui, share_count)
+        self.idle_add(self._process_hello_ui, ss, c, auth_caps, send_ui, share_count)
+
+    def get_client_connection_class(self):
+        from xpra.server.source.client_connection import ClientConnection
+        return ClientConnection
 
 
-    def parse_hello_ui(self, ss, c, auth_caps, send_ui, share_count):
+    def _process_hello_ui(self, ss, c, auth_caps, send_ui, share_count):
         #adds try:except around parse hello ui code:
         try:
             if self._closing:
                 raise Exception("server is shutting down")
-            self.do_parse_hello_ui(ss, c, auth_caps, send_ui, share_count)
+
+            self.parse_hello(ss, c, send_ui)
+            #send_hello will take care of sending the current and max screen resolutions
+            root_w, root_h = self.get_root_window_size()
+            self.send_hello(ss, root_w, root_h, auth_caps)
+            self.add_new_client(ss, c, send_ui, share_count)
+            self.send_initial_data(ss, c, send_ui, share_count)
+            self.client_startup_complete(ss)
+
             if self._closing:
                 raise Exception("server is shutting down")
         except Exception as e:
             #log exception but don't disclose internal details to the client
             p = ss.protocol
-            log("parse_hello_ui%s", (ss, c, auth_caps, send_ui, share_count), exc_info=True)
+            log("_process_hello_ui%s", (ss, c, auth_caps, send_ui, share_count), exc_info=True)
             log.error("Error: processing new connection from %s:", p or ss)
             log.error(" %s", e)
             if p:
                 self.disconnect_client(p, SERVER_ERROR, "error accepting new connection")
 
-    def do_parse_hello_ui(self, ss, c, auth_caps, send_ui, share_count):
-        #process screen size (if needed)
-        if send_ui:
-            for bc in SERVER_BASES:
-                if bc!=ServerCore:
-                    bc.parse_hello(self, ss, c, send_ui)
-        root_w, root_h = self.get_root_window_size()
+    def parse_hello(self, ss, c, send_ui):
+        for bc in SERVER_BASES:
+            if bc!=ServerCore:
+                bc.parse_hello(self, ss, c, send_ui)
 
-        #send_hello will take care of sending the current and max screen resolutions
-        self.send_hello(ss, root_w, root_h, auth_caps)
-
-        if send_ui and server_features.windows:
-            self.send_initial_windows(ss, share_count>0)
-            self.send_initial_cursors(ss, share_count>0)
-        self.client_startup_complete(ss)
+    def add_new_client(self, ss, c, send_ui, share_count):
+        for bc in SERVER_BASES:
+            if bc!=ServerCore:
+                bc.add_new_client(self, ss, c, send_ui, share_count)
 
+    def send_initial_data(self, ss, c, send_ui, share_count):
+        for bc in SERVER_BASES:
+            if bc!=ServerCore:
+                bc.send_initial_data(self, ss, c, send_ui, share_count)
 
     def client_startup_complete(self, ss):
         ss.startup_complete()
@@ -478,7 +503,7 @@ class ServerBase(ServerBaseClass):
         #if len(packet>=2):
         #    uuid = packet[1]
         if len(packet)>=4:
-            categories = packet[3]
+            categories = tuple(bytestostr(x) for x in packet[3])
         def info_callback(_proto, info):
             assert proto==_proto
             if categories:
@@ -503,7 +528,7 @@ class ServerBase(ServerBaseClass):
             try:
                 merge_dicts(info, c.get_ui_info(self, proto, client_uuids, *args))
             except Exception:
-                log.error("Error gathering UI info", exc_info=True)
+                log.error("Error gathering UI info on %s", c, exc_info=True)
         return info
 
     def get_thread_info(self, proto):
@@ -514,9 +539,6 @@ class ServerBase(ServerBaseClass):
         log("ServerBase.get_info%s", (proto, client_uuids))
         start = monotonic_time()
         info = ServerCore.get_info(self, proto)
-        server_info = info.setdefault("server", {})
-        if self.mem_bytes:
-            server_info["total-memory"] = self.mem_bytes
         if client_uuids:
             sources = [ss for ss in self._server_sources.values() if ss.uuid in client_uuids]
         else:
@@ -612,6 +634,7 @@ class ServerBase(ServerBaseClass):
                 if v is None:
                     log.warn("removing invalid None property for %s", k)
                     continue
+                k = strtobytes(k)
                 if not k.startswith(b"encoding"):
                     ncp[k] = v
             if ncp:
@@ -758,8 +781,8 @@ class ServerBase(ServerBaseClass):
         #this ensures that from now on we ignore any incoming packets coming
         #from this connection as these could potentially set some keys pressed, etc
         try:
-            del self._potential_protocols[protocol]
-        except:
+            self._potential_protocols.remove(protocol)
+        except ValueError:
             pass
         source = self._server_sources.pop(protocol, None)
         if source:
@@ -795,9 +818,9 @@ class ServerBase(ServerBaseClass):
             for c in SERVER_BASES:
                 if c!=ServerCore:
                     try:
-                        c.reset_state(self)
+                        c.last_client_exited(self)
                     except:
-                        log("last_client_exited calling %s", c.reset_state, exc_info=True)
+                        log("last_client_exited calling %s", c.last_client_exited, exc_info=True)
 
 
     def set_ui_driver(self, source):
@@ -894,11 +917,14 @@ class ServerBase(ServerBaseClass):
                 ss = self._server_sources.get(proto)
                 if not self._closing and not proto._closed and (ss is None or not ss.is_closed()):
                     netlog("invalid packet: %s", packet)
-                    netlog.error("unknown or invalid packet type: %s from %s", packet_type, proto)
+                    netlog.error("Error: unknown or invalid packet type '%s'", packet_type)
+                    netlog.error(" received from %s", proto)
                 if not ss:
                     proto.close()
             self.idle_add(invalid_packet)
         except KeyboardInterrupt:
             raise
-        except:
-            netlog.error("Unhandled error while processing a '%s' packet from peer using %s", packet_type, handler, exc_info=True)
+        except Exception:
+            netlog.error("Error processing a '%s' packet", packet_type)
+            netlog.error(" received from %s:", proto)
+            netlog.error(" using %s", handler, exc_info=True)
