@@ -1,5 +1,5 @@
 # This file is part of Xpra.
-# Copyright (C) 2017 Antoine Martin <antoine@devloop.org.uk>
+# Copyright (C) 2017 Antoine Martin <antoine@xpra.org>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
@@ -7,11 +7,10 @@ import os.path
 import socket
 
 from xpra.scripts.config import InitException
-from xpra.os_util import getuid, get_username_for_uid, get_groups, get_group_id, path_permission_info, monotonic_time, WIN32, OSX, POSIX
+from xpra.os_util import getuid, get_username_for_uid, get_groups, get_group_id, path_permission_info, monotonic_time, umask_context, WIN32, OSX, POSIX
 from xpra.util import envint, envbool, csv, DEFAULT_PORT
 from xpra.platform.dotxpra import DotXpra, norm_makepath
 
-
 #what timeout value to use on the socket probe attempt:
 WAIT_PROBE_TIMEOUT = envint("XPRA_WAIT_PROBE_TIMEOUT", 6)
 GROUP = os.environ.get("XPRA_GROUP", "xpra")
@@ -33,15 +32,12 @@ def get_network_logger():
 
 def create_unix_domain_socket(sockpath, socket_permissions=0o600):
     #convert this to a umask!
-    umask = 0o777-socket_permissions
+    umask = (0o777-socket_permissions) & 0o777
     listener = socket.socket(socket.AF_UNIX)
     listener.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
     #bind the socket, using umask to set the correct permissions
-    orig_umask = os.umask(umask)
-    try:
+    with umask_context(umask):
         listener.bind(sockpath)
-    finally:
-        os.umask(orig_umask)
     try:
         inode = os.stat(sockpath).st_ino
     except:
@@ -132,6 +128,9 @@ def setup_tcp_socket(host, iport, socktype="tcp"):
         except:
             pass
     add_cleanup(cleanup_tcp_socket)
+    if iport==0:
+        iport = tcp_socket.getsockname()[1]
+        log.info("allocated %s port %i on %s", socktype, iport, host)
     log("%s: %s:%s : %s", socktype, host, iport, socket)
     return socktype, tcp_socket, (host, iport)
 
@@ -161,6 +160,9 @@ def setup_udp_socket(host, iport, socktype="udp"):
         except:
             pass
     add_cleanup(cleanup_udp_socket)
+    if iport==0:
+        iport = udp_socket.getsockname()[1]
+        log.info("allocated UDP port %i for %s", iport, host)
     log("%s: %s:%s : %s", socktype, host, iport, socket)
     return socktype, udp_socket, (host, iport)
 
@@ -176,6 +178,8 @@ def parse_bind_ip(bind_ip, default_port=DEFAULT_PORT):
                 host = "127.0.0.1"
             if not port:
                 iport = default_port
+            elif port=="0":
+                iport = 0
             else:
                 try:
                     iport = int(port)
@@ -427,21 +431,27 @@ def mdns_publish(display_name, mode, listen_on, text_dict={}):
     global MDNS_WARNING
     if MDNS_WARNING is True:
         return
-    PREFER_PYBONJOUR = envbool("XPRA_PREFER_PYBONJOUR", False) or WIN32 or OSX
     try:
         from xpra.net import mdns
         assert mdns
+        from xpra.net.mdns import XPRA_MDNS_TYPE, RFB_MDNS_TYPE
+        PREFER_PYBONJOUR = envbool("XPRA_PREFER_PYBONJOUR", False) or WIN32 or OSX
+        PREFER_ZEROCONF = envbool("XPRA_PREFER_ZEROCONF", False)
         if PREFER_PYBONJOUR:
             from xpra.net.mdns.pybonjour_publisher import BonjourPublishers as MDNSPublishers, get_interface_index
+        elif PREFER_ZEROCONF:
+            from xpra.net.mdns.zeroconf_publisher import ZeroconfPublishers as MDNSPublishers, get_interface_index
         else:
             from xpra.net.mdns.avahi_publisher import AvahiPublishers as MDNSPublishers, get_interface_index
     except ImportError as e:
         MDNS_WARNING = True
-        log = get_network_logger()
+        from xpra.log import Logger
+        log = Logger("mdns")
         log("mdns import failure", exc_info=True)
-        log.warn("Warning: failed to load the mdns %s publisher:", ["avahi", "pybonjour"][PREFER_PYBONJOUR])
+        log.warn("Warning: failed to load the mdns publisher")
         log.warn(" %s", e)
-        log.warn(" either fix your installation or use the 'mdns=no' option")
+        log.warn(" either install the 'python-avahi' module")
+        log.warn(" or use the 'mdns=no' option")
         return
     d = text_dict.copy()
     d["mode"] = mode
@@ -455,9 +465,10 @@ def mdns_publish(display_name, mode, listen_on, text_dict={}):
         name = "Xpra"
     if display_name and not (OSX or WIN32):
         name += " %s" % display_name
-    if mode!="tcp":
+    if mode not in ("tcp", "rfb"):
         name += " (%s)" % mode
-    ap = MDNSPublishers(f_listen_on.values(), name, text_dict=d)
+    service_type = {"rfb" : RFB_MDNS_TYPE}.get(mode, XPRA_MDNS_TYPE)
+    ap = MDNSPublishers(f_listen_on.values(), name, service_type=service_type, text_dict=d)
     from xpra.scripts.server import add_when_ready, add_cleanup
     add_when_ready(ap.start)
     add_cleanup(ap.stop)
