@@ -1,47 +1,52 @@
 # -*- coding: utf-8 -*-
 # This file is part of Xpra.
-# Copyright (C) 2016-2018 Antoine Martin <antoine@devloop.org.uk>
+# Copyright (C) 2016-2019 Antoine Martin <antoine@xpra.org>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
 import os
-from gtk import gdk
-import gobject
 import socket
 
-from xpra.os_util import get_generic_os_name, PYTHON3
-if PYTHON3:
-    raise ImportError("desktop server needs porting to python3 / gtk3")
+from xpra.os_util import get_generic_os_name, load_binary_file, PYTHON3
 from xpra.util import updict, log_screen_sizes
-from xpra.platform.paths import get_icon
+from xpra.platform.paths import get_icon, get_icon_filename
 from xpra.platform.gui import get_wm_name
 from xpra.server import server_features
 from xpra.gtk_common.gobject_util import one_arg_signal, no_arg_signal
 from xpra.gtk_common.gobject_compat import import_glib
-from xpra.gtk_common.error import xswallow
-from xpra.gtk_common.gtk_util import get_screen_sizes, get_root_size, get_xwindow
-from xpra.x11.gtk2.models.model_stub import WindowModelStub
-from xpra.x11.gtk2.gdk_bindings import (
-                               add_catchall_receiver,       #@UnresolvedImport
-                               remove_catchall_receiver,    #@UnresolvedImport
-                               add_event_receiver,          #@UnresolvedImport
-                               init_x11_filter,             #@UnresolvedImport
-                               cleanup_x11_filter,          #@UnresolvedImport
-                               cleanup_all_event_receivers  #@UnresolvedImport
-                               )
+from xpra.gtk_common.error import xswallow, XError
+from xpra.gtk_common.gtk_util import (
+    get_screen_sizes, get_root_size,
+    get_xwindow,
+    display_get_default,
+    PARAM_READABLE, PARAM_READWRITE,
+    )
+from xpra.x11.models.model_stub import WindowModelStub
+from xpra.x11.gtk_x11.gdk_bindings import (
+    add_catchall_receiver, remove_catchall_receiver,
+    add_event_receiver,          #@UnresolvedImport
+    init_x11_filter, cleanup_x11_filter,          #@UnresolvedImport
+    cleanup_all_event_receivers  #@UnresolvedImport
+   )
 from xpra.x11.bindings.window_bindings import X11WindowBindings #@UnresolvedImport
 from xpra.x11.xroot_props import XRootPropWatcher
-from xpra.x11.gtk2.window_damage import WindowDamageHandler
-X11Window = X11WindowBindings()
+from xpra.x11.gtk_x11.window_damage import WindowDamageHandler
 from xpra.x11.bindings.keyboard_bindings import X11KeyboardBindings #@UnresolvedImport
-X11Keyboard = X11KeyboardBindings()
 from xpra.x11.bindings.randr_bindings import RandRBindings #@UnresolvedImport
-RandR = RandRBindings()
 from xpra.x11.x11_server_base import X11ServerBase, mouselog
 from xpra.gtk_common.error import xsync
-
+from xpra.gtk_common.gobject_compat import import_gobject
 from xpra.log import Logger
+
 log = Logger("server")
+
+gobject = import_gobject()
+glib = import_glib()
+
+X11Window = X11WindowBindings()
+X11Keyboard = X11KeyboardBindings()
+RandR = RandRBindings()
+
 windowlog = Logger("server", "window")
 geomlog = Logger("server", "window", "geometry")
 settingslog = Logger("x11", "xsettings")
@@ -49,8 +54,6 @@ metadatalog = Logger("x11", "metadata")
 screenlog = Logger("screen")
 iconlog = Logger("icon")
 
-glib = import_glib()
-
 
 class DesktopModel(WindowModelStub, WindowDamageHandler):
     __gsignals__ = {}
@@ -64,25 +67,29 @@ class DesktopModel(WindowModelStub, WindowDamageHandler):
         "iconic": (gobject.TYPE_BOOLEAN,
                    "ICCCM 'iconic' state -- any sort of 'not on desktop'.", "",
                    False,
-                   gobject.PARAM_READWRITE),
+                   PARAM_READWRITE),
         "focused": (gobject.TYPE_BOOLEAN,
                        "Is the window focused", "",
                        False,
-                       gobject.PARAM_READWRITE),
+                       PARAM_READWRITE),
         "size-hints": (gobject.TYPE_PYOBJECT,
                        "Client hints on constraining its size", "",
-                       gobject.PARAM_READABLE),
+                       PARAM_READABLE),
         "wm-name": (gobject.TYPE_PYOBJECT,
                        "The name of the window manager or session manager", "",
-                       gobject.PARAM_READABLE),
-        "icon": (gobject.TYPE_PYOBJECT,
+                       PARAM_READABLE),
+        "icons": (gobject.TYPE_PYOBJECT,
                        "The icon of the window manager or session manager", "",
-                       gobject.PARAM_READABLE),
+                       PARAM_READABLE),
         }
 
 
-    _property_names         = ["xid", "client-machine", "window-type", "shadow", "size-hints", "class-instance", "focused", "title", "depth", "icon"]
-    _dynamic_property_names = ["size-hints", "title", "icon"]
+    _property_names         = [
+        "xid", "client-machine", "window-type",
+        "shadow", "size-hints", "class-instance",
+        "focused", "title", "depth", "icons",
+        ]
+    _dynamic_property_names = ["size-hints", "title", "icons"]
 
     def __init__(self, root, resize_exact=False):
         WindowDamageHandler.__init__(self, root)
@@ -101,7 +108,7 @@ class DesktopModel(WindowModelStub, WindowDamageHandler):
         screen = self.client_window.get_screen()
         screen.connect("size-changed", self._screen_size_changed)
         self.update_size_hints(screen)
-        self._depth = X11Window.get_depth(self.client_window.xid)
+        self._depth = X11Window.get_depth(get_xwindow(self.client_window))
         self._managed = True
         self._setup_done = True
 
@@ -129,14 +136,21 @@ class DesktopModel(WindowModelStub, WindowDamageHandler):
         return self._updateprop("wm-name", wm_name)
 
     def update_icon(self):
-        icon = None
+        icons = None
         try:
-            icon_name = (get_wm_name() or "").lower()+".png"
-            icon = get_icon(icon_name)
-            iconlog("get_icon(%s)=%s", icon_name, icon)
+            icon_name = get_icon_filename((get_wm_name() or "").lower()+".png")
+            from PIL import Image
+            img = Image.open(icon_name)
+            iconlog("Image(%s)=%s", icon_name, img)
+            if img:
+                icon_data = load_binary_file(icon_name)
+                assert icon_data
+                w, h = img.size
+                icon = (w, h, "png", icon_data)
+                icons = (icon,)
         except:
-            iconlog("failed to return window icon")
-        return self._updateprop("icon", icon)
+            iconlog("failed to return window icon", exc_info=True)
+        return self._updateprop("icons", icons)
 
 
     def get_geometry(self):
@@ -149,28 +163,30 @@ class DesktopModel(WindowModelStub, WindowDamageHandler):
         return bool(self._xshm_handle)
 
 
-    def get_default_window_icon(self):
+    def get_default_window_icon(self, _size):
         icon_name = get_generic_os_name()+".png"
-        return get_icon(icon_name)
+        icon = get_icon(icon_name)
+        if not icon:
+            return None
+        return icon.get_width(), icon.get_height(), "BGRA", icon.get_pixels()
 
 
     def get_property(self, prop):
         if prop=="xid":
-            return int(self.client_window.xid)
-        elif prop=="depth":
+            return int(get_xwindow(self.client_window))
+        if prop=="depth":
             return self._depth
-        elif prop=="title":
+        if prop=="title":
             return get_wm_name() or "xpra desktop"
-        elif prop=="client-machine":
+        if prop=="client-machine":
             return socket.gethostname()
-        elif prop=="window-type":
+        if prop=="window-type":
             return ["NORMAL"]
-        elif prop=="shadow":
+        if prop=="shadow":
             return True
-        elif prop=="class-instance":
+        if prop=="class-instance":
             return ("xpra-desktop", "Xpra-Desktop")
-        else:
-            return gobject.GObject.get_property(self, prop)
+        return gobject.GObject.get_property(self, prop)
 
     def _screen_size_changed(self, screen):
         w, h = screen.get_width(), screen.get_height()
@@ -200,10 +216,10 @@ class DesktopModel(WindowModelStub, WindowDamageHandler):
                 try:
                     with xsync:
                         screen_sizes = RandR.get_xrr_screen_sizes()
-                except:
+                except XError:
                     screenlog("failed to query screen sizes", exc_info=True)
                 else:
-                    if len(screen_sizes)==0:
+                    if not screen_sizes:
                         use_fixed_size()
                     else:
                         #find the maximum size supported:
@@ -288,7 +304,7 @@ class XpraDesktopServer(DesktopServerBaseClass):
     def x11_init(self):
         X11ServerBase.x11_init(self)
         assert init_x11_filter() is True
-        display = gdk.display_get_default()
+        display = display_get_default()
         screens = display.get_n_screens()
         for n in range(screens):
             screen = display.get_screen(n)
@@ -328,24 +344,26 @@ class XpraDesktopServer(DesktopServerBaseClass):
 
     def configure_best_screen_size(self):
         """ for the first client, honour desktop_mode_size if set """
-        root_w, root_h = self.root_window.get_size()
+        root_w, root_h = self.root_window.get_geometry()[2:4]
         if not self.randr:
             screenlog("configure_best_screen_size() no randr")
             return root_w, root_h
-        sss = self._server_sources.values()
+        sss = tuple(self._server_sources.values())
         if len(sss)!=1:
             screenlog.info("screen used by %i clients:", len(sss))
             return root_w, root_h
-        requested_size = sss[0].desktop_mode_size
+        ss = sss[0]
+        requested_size = ss.desktop_mode_size
         if not requested_size:
             screenlog("configure_best_screen_size() client did not request a specific desktop mode size")
             return root_w, root_h
         w, h = requested_size
-        screenlog("client requested desktop mode resolution is %sx%s (current server resolution is %sx%s)", w, h, root_w, root_h)
+        screenlog("client requested desktop mode resolution is %sx%s (current server resolution is %sx%s)",
+                  w, h, root_w, root_h)
         if w<=0 or h<=0 or w>=32768 or h>=32768:
             screenlog("configure_best_screen_size() client requested an invalid desktop mode size: %s", requested_size)
             return root_w, root_h
-        return self.set_screen_size(w, h)
+        return self.set_screen_size(w, h, ss.screen_resize_bigger)
 
     def cancel_resize_timer(self):
         rt = self.resize_timer
@@ -390,7 +408,9 @@ class XpraDesktopServer(DesktopServerBaseClass):
 
 
     def get_server_mode(self):
-        return "X11 desktop"
+        if PYTHON3:
+            return "GTK3 X11 desktop"
+        return "GTK2 X11 desktop"
 
     def make_hello(self, source):
         capabilities = X11ServerBase.make_hello(self, source)
@@ -412,7 +432,7 @@ class XpraDesktopServer(DesktopServerBaseClass):
     def load_existing_windows(self):
         #at present, just one  window is forwarded:
         #the root window covering the whole display
-        display = gdk.display_get_default()
+        display = display_get_default()
         screens = display.get_n_screens()
         with xsync:
             for n in range(screens):
@@ -455,7 +475,7 @@ class XpraDesktopServer(DesktopServerBaseClass):
 
     def _contents_changed(self, window, event):
         log("contents changed on %s: %s", window, event)
-        self._damage(window, event.x, event.y, event.width, event.height)
+        self.refresh_window_area(window, event.x, event.y, event.width, event.height)
 
 
     def _set_window_state(self, proto, wid, window, new_window_state):
@@ -496,7 +516,7 @@ class XpraDesktopServer(DesktopServerBaseClass):
             self._set_window_state(proto, wid, window, packet[7])
         if len(packet)>=7:
             self._set_client_properties(proto, wid, window, packet[6])
-        self._damage(window, 0, 0, w, h)
+        self.refresh_window_area(window, 0, 0, w, h)
 
 
     def _process_unmap_window(self, proto, packet):
@@ -531,8 +551,7 @@ class XpraDesktopServer(DesktopServerBaseClass):
             return
         damage = False
         if len(packet)>=9:
-            changes = self._set_window_state(proto, wid, window, packet[8])
-            damage = len(changes)>0
+            damage = bool(self._set_window_state(proto, wid, window, packet[8]))
         if not skip_geometry:
             owx, owy, oww, owh = window.get_geometry()
             geomlog("_process_configure_window(%s) old window geometry: %s", packet[1:], (owx, owy, oww, owh))
@@ -545,7 +564,7 @@ class XpraDesktopServer(DesktopServerBaseClass):
                 self._set_client_properties(proto, wid, window, cprops)
         self._window_mapped_at(proto, wid, window, (x, y, w, h))
         if damage:
-            self._damage(window, 0, 0, w, h)
+            self.refresh_window_area(window, 0, 0, w, h)
 
 
     def _adjust_pointer(self, proto, wid, pointer):
@@ -556,7 +575,7 @@ class XpraDesktopServer(DesktopServerBaseClass):
         pointer = super(XpraDesktopServer, self)._adjust_pointer(proto, wid, pointer)
         #maybe the pointer is off-screen:
         ww, wh = window.get_dimensions()
-        x, y = pointer
+        x, y = pointer[:2]
         if x<0 or x>=ww or y<0 or y>=wh:
             self.suspend_cursor(proto)
             return None
